/*	$NecBSD: autoscan.c,v 1.48.4.1.2.4 2000/07/21 07:50:26 honda Exp $	*/
/*	$NetBSD$	*/

/*
 * [NetBSD for NEC PC-98 series]
 *  Copyright (c) 1996, 1997, 1998
 *	NetBSD/pc98 porting staff. All rights reserved.
 * 
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. The name of the author may not be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * Copyright (c) 1994, 1995, 1996, 1997, 1998, 2000
 *	Naofumi HONDA.  All rights reserved.
 */

#include "opt_autoscan.h"
#include <sys/param.h>
#include <sys/systm.h>
#include <sys/buf.h>
#include <sys/disklabel.h>
#include <sys/syslog.h>
#include <sys/device.h>
#include <sys/malloc.h>
#include <machine/dvcfg.h>
#include <machine/sysinfo.h>
#include <i386/i386/autoscan.h>
#include "sd.h"
#include "wd.h"

/*******************************************************************
 * AUTOSCAN
 *******************************************************************/
int autoscan_debug;

static int find_pc98_dospart __P((struct dos_check_args *, int));
static int find_at_dospart __P((struct dos_check_args *, int));
static void change_disk_geometry __P((struct dos_check_args *));
static struct part_type *check_pc98_disktype __P((sysid_t));
static struct part_type *check_at_disktype __P((sysid_t));
static int scan_geo_by_config __P((u_char *, u_int *, u_int *));
static int lookup_bootinfo_disk_geometry __P((dev_t, struct disklabel *, struct cpu_disklabel *));

#define	b_cylin	b_resid

/*******************************************************************
 * PARTITION TYPE
 *******************************************************************/
#ifndef	DOSSID_LINUX
#define	DOSSID_LINUX	0xe2
#endif

struct part_type {
	sysid_t	sysid;		/* disk system id */
	sysid_t mask;		/* identification mask */
	int fstype;		/* file system type */
#define	ROK	0x01
#define	FOK	0x02
	u_char flags;		/* flags */
};

#define	MAKEID(mid, sid) sysid_activate(sysid_make_id((mid), (sid)))
struct part_type part_pc98_types[] = {
	{MAKEID(0x14, DOSSID_NETBSD), 0xff7f, FS_BSDFFS, ROK},
	{MAKEID(0x21, 0xb1), 	      0xff7f, FS_HPFS,   ROK | FOK},
	{MAKEID(0x11, DOSSID_MSDOS),  0x8f7f, FS_MSDOS,  ROK | FOK},
	{MAKEID(0x20, DOSSID_MSDOS),  0x8f7f, FS_MSDOS,  ROK | FOK},
	{MAKEID(0x21, DOSSID_MSDOS),  0x8f7f, FS_MSDOS,  ROK | FOK},
	{MAKEID(0x22, DOSSID_MSDOS),  0x8f7f, FS_MSDOS,  ROK | FOK},
	{MAKEID(0x23, DOSSID_MSDOS),  0x8f7f, FS_MSDOS,  ROK | FOK},
	{MAKEID(0x07, DOSSID_MSDOS),  0x8f7f, FS_HPFS,   ROK | FOK},
	{MAKEID(0x75, DOSSID_LINUX),  0xff7f, FS_EX2FS,	 ROK | FOK},
	{MAKEID(0xa0, DOSSID_LINUX),  0xff7f, FS_EX2FS,  ROK | FOK},
	{0},
};

struct part_type part_at_types[] = {
	{DOSPTYP_FAT12,  0xff, FS_MSDOS, ROK | FOK},
	{DOSPTYP_FAT16S, 0xff, FS_MSDOS, ROK | FOK},
	{DOSPTYP_FAT16B, 0xff, FS_MSDOS, ROK | FOK},
	{DOSPTYP_FAT32,  0xff, FS_MSDOS, ROK | FOK},
	{DOSPTYP_FAT32L, 0xff, FS_MSDOS, ROK | FOK},
	{DOSPTYP_FAT16L, 0xff, FS_MSDOS, ROK | FOK},
	{DOSPTYP_LNXEXT2, 0xff, FS_EX2FS,  ROK | FOK},
	{SYSID_AT_386BSD, 0xff, FS_BSDFFS, ROK},
	{SYSID_AT_NETBSD, 0xff, FS_BSDFFS, ROK},
	{0},
};

/*******************************************************************
 * EXPORT FUNCS
 *******************************************************************/
int
read_cpudisklabel(dosp)
	struct dos_check_args *dosp;
{
	struct cpu_disklabel *cdp = dosp->cdp;
	struct disklabel *lp = dosp->lp;
	int size, error = 0;
	struct buf *bp;

	cdp->cd_nslices = 0;
	cdp->cd_nothers = 0;
	cdp->cd_nours = 0;

	size = ((int) lp->d_secsize) < NBPG ? NBPG : lp->d_secsize;
	bp = geteblk(size);
	bp->b_dev = dosp->dev;
	bp->b_blkno = DOSBBSECTOR;
	bp->b_cylin = DOSBBSECTOR / lp->d_secpercyl;
	bp->b_bcount = size;
	bp->b_flags = B_BUSY | B_READ;

	(*dosp->start) (bp);
	error = biowait(bp);
	if (error != 0)
	{
		error = EIO;
		goto out;
	}

	dosp->bufp = bp->b_data;
	error = setup_cpudisklabel(dosp);

out:
	bp->b_flags |= B_INVAL;
	brelse(bp);
	return error;
}

int
setup_cpudisklabel(dosp)
	struct dos_check_args *dosp;
{

	if (find_pc98_dospart(dosp, 0) > 0)
		return 0;

	/* II: check a IBM dos partition table */
	if (find_at_dospart(dosp, 0) > 0)
		return 0;

	(void) find_pc98_dospart(dosp, 1);
	return ENOENT;
}

/*******************************************************************
 * Disk type functions
 *******************************************************************/
static struct part_type *check_disktype __P((sysid_t, struct part_type *));

static struct part_type *
check_pc98_disktype(id)
	sysid_t id;
{

	return check_disktype(id, &part_pc98_types[0]);
}

static struct part_type *
check_at_disktype(id)
	sysid_t id;
{

	return check_disktype(id, &part_at_types[0]);
}

static struct part_type *
check_disktype(id, pt)
	sysid_t id;
	struct part_type *pt;
{

	for ( ; pt->sysid != 0; pt ++)
	{
		if ((id & pt->mask) == (pt->sysid & pt->mask) &&
		    (pt->flags & ROK) != 0)
			return pt;
	}
	return NULL;
}

/*******************************************************************
 * CHECK PC-98 & IBM DOS PARTITIONS
 *******************************************************************/
static int
find_at_dospart(dosp, force)
	struct dos_check_args *dosp;
	int force;
{
	struct disklabel *lp = dosp->lp;
	struct cpu_disklabel *cdp = dosp->cdp;
	struct at_dos_partition *dp;
	int slno, i, nfound = 0;
	char *cbp;


	/* (I)
	 * Are there NetBSD or AT MSDOS entries ?
	 */
 	dp = (struct at_dos_partition *) (dosp->bufp + AT_DOSPARTOFF);
	for (i = 0; i < AT_NDOSPART; i++, dp++)
	{
		sysid_t id;

		if (dp->dp_size == 0)
			continue;

		id = dos_partition_at_read_sysid(dp);
		if (check_at_disktype(id) != NULL)
			nfound ++;
	}

	cdp->cd_cksum = 0;
	cbp = (char *) dosp->bufp;
	for (i = 0; i < lp->d_secsize; i ++)
	{
		cdp->cd_cksum += cbp[i];
	}

	if (nfound == 0)
		return nfound;

	/* (II)
	 * Load info
	 */
 	dp = (struct at_dos_partition *) (dosp->bufp + AT_DOSPARTOFF);
	for (i = slno = 0; i < AT_NDOSPART; i++, dp++)
	{
		struct part_type *ptype;
		sysid_t id;

		if (dp->dp_size == 0)
			continue;
		id = dos_partition_at_read_sysid(dp);
		if ((ptype = check_at_disktype(id)) == NULL)
			continue;

		cdp->cd_se[slno].se_start = dp->dp_start;
		cdp->cd_se[slno].se_size = dp->dp_size;
		cdp->cd_se[slno].se_fstype = ptype->fstype;
		if (IS_BSDFS(&cdp->cd_se[slno]) != 0)
		{
			cdp->cd_nours ++;
		}
		else
		{
			cdp->cd_nothers ++;
		}
		slno ++;
	}
	cdp->cd_nslices = slno;

	return nfound;
}

static int
find_pc98_dospart(dosp, force)
	struct dos_check_args *dosp;
	int force;
{
	struct disklabel *lp = dosp->lp;
	struct cpu_disklabel *cdp = dosp->cdp;
	struct dos_partition *dp;
	int ndospart, slno, i, nfound = 0;
	sysid_t id;
	char *cbp;

	/* (I)
	 * Are there NetBSD or PC98 msdos entries ?
	 */
	ndospart = (lp->d_secsize / 32);
	if (ndospart > NSLICE_ENTRY)
		ndospart = NSLICE_ENTRY;

 	dp = (struct dos_partition *) (dosp->bufp + lp->d_secsize);
	for (i = 0; i < ndospart; i++, dp++)
	{
		id = dos_partition_pc98_read_sysid(dp);
		if (id == (sysid_t) 0)
			break;
		if (check_pc98_disktype(id) != NULL)
			nfound ++;
	}

	cdp->cd_cksum = 0;
	cbp = (char *) (dosp->bufp + lp->d_secsize);
	for (i = 0; i < lp->d_secsize; i ++)
	{
		cdp->cd_cksum += cbp[i];
	}

	if (nfound == 0 && force == 0)
		return nfound;

	/* (II)
	 * Yes. Update the geometry.
	 */
	if (dosp->flag & GEO_CHG_EN)
		change_disk_geometry(dosp);

	if (nfound == 0)
		return nfound;

	/* (III)
	 * Load info
	 */
 	dp = (struct dos_partition *) (dosp->bufp + lp->d_secsize);
	for (i = slno = 0; i < ndospart; i++, dp++)
	{
		struct part_type *ptype;
		sysid_t id;

		if ((id = dos_partition_pc98_read_sysid(dp)) == 0)
			break;
		if ((ptype = check_pc98_disktype(id)) == NULL)
			continue;

		cdp->cd_se[slno].se_start = disk_start_sector(lp, dp);
		cdp->cd_se[slno].se_size = disk_size_sector(lp, dp);
		cdp->cd_se[slno].se_fstype = ptype->fstype;
		if (IS_BSDFS(&cdp->cd_se[slno]) != 0)
		{
			cdp->cd_nours ++;
		}
		else
		{
			cdp->cd_nothers ++;
		}
		slno ++;
	}
	cdp->cd_nslices = slno;

	return nfound;
}

/*******************************************************************
 * GEOMETRY SCANNING (MAGIC)
 *******************************************************************/
#include <i386/Cbus/dev/magicvar.h>

int cfd_idx;
struct config_device_geometry cfd_data[CFD_NGEOMETRIES + 1];

int
scan_geo_by_config(name, tracks, sectors)
	u_char *name;
	u_int *tracks, *sectors;
{
	struct config_device_geometry *cfdp;
	int no;

	for (cfdp = cfd_data, no = 0; no < CFD_NGEOMETRIES; cfdp ++, no ++)
	{
		if (cfdp->cfd_name[0] == 0)
			continue;

		if(strncmp(name, cfdp->cfd_name, CFD_NAMELEN) == 0 &&
		   cfdp->cfd_tracks != 0 && cfdp->cfd_sectors != 0)
		{
			*tracks = cfdp->cfd_tracks;
			*sectors = cfdp->cfd_sectors;
			return 0;
		}
	}

	return ENOENT;
}
			
/*******************************************************************
 * SCSI GEOMETRY SCANNING
 *******************************************************************/
#if	NSD_ATAPIBUS > 0 || NSD_SCSIBUS > 0
#include <sys/disk.h>
#include <dev/scsipi/scsi_all.h>
#include <dev/scsipi/scsipi_all.h>
#include <dev/scsipi/scsi_disk.h>
#include <dev/scsipi/scsiconf.h>
#include <i386/conf/sdgsets.h>
#include <dev/scsipi/sdvar.h>

#define	b_cylin	b_resid
#define	OS_ANY	0

#if	NSD_SCSIBUS > 0
#ifdef	SCSIBIOS_SCSIBUS_ID
int scsibios_scsibus_id = SCSIBIOS_SCSIBUS_ID;
#else	/* !SCSIBIOS_SCSIBUS_ID */
int scsibios_scsibus_id = 0;
#endif	/* !SCSIBIOS_SCSIBUS_ID */

static struct sd_softc *snoop_sd_softc __P((int));
static int snoop_sd_dvcfg __P((int, u_int *));
static int snoop_scsibus_info __P((int, u_int *, u_int *, u_int *));

static int scsi_geo_match __P((struct dos_check_args *, u_int, u_int, int));
static int scan_geo_by_ur __P((struct dos_check_args *, u_int *, u_int *));
static int scan_geo_by_tl __P((struct dos_check_args *, u_int, u_int *, u_int *));
static u_int round_up __P((u_int));
static u_char sdsymbol[4] = {'s', 'd', '0', 0};

static struct sd_softc *
snoop_sd_softc(unit)
	int unit;
{
	extern struct cfdriver sd_cd;

	if (unit >= sd_cd.cd_ndevs)
		return NULL;

	return ((struct sd_softc *) (sd_cd.cd_devs[unit]));
}

static int
snoop_scsibus_info(unit, busid, id, lun) 
	int unit;
	u_int *busid, *id, *lun;
{
	struct sd_softc *sc;
	struct scsipi_link *sl;

	if ((sc = snoop_sd_softc(unit)) == NULL)
		return ENOENT;

	if ((sl = sc->sc_link) == NULL)
		return ENOENT;	

	*busid = (u_int) sl->scsipi_scsi.scsibus;
	*id = (u_int) sl->scsipi_scsi.target;
	*lun = (u_int) sl->scsipi_scsi.lun;
	return 0;
}

static int
snoop_sd_dvcfg(unit, dvcfg)
	int unit;
	u_int *dvcfg;
{
	struct sd_softc *sc;

	*dvcfg = 0;
	if ((sc = snoop_sd_softc(unit)) == NULL)
		return ENOENT;

	if (sc->sc_dev.dv_cfdata == NULL)
		return ENOENT;

	*dvcfg = sc->sc_dev.dv_cfdata->cf_flags;
	return 0;
}

static u_int
round_up(val)
	u_int val;
{
	u_int eval;

	for(eval = 1; val < eval || val >= eval * 2; eval <<= 1)
		;

	return eval;
}

static int
scsi_geo_match(dosp, tracks, sectors, ostype)
	struct dos_check_args *dosp;
	u_int tracks, sectors;
	int ostype;
{
	struct disklabel *lp = dosp->lp;
	struct disklabel slp;
	struct buf *tbp;
	u_char *buf;
	u_int total;
	int error = EINVAL;

	slp = *lp;

	total = lp->d_secperunit;
	lp->d_ntracks = tracks;
	lp->d_nsectors = sectors;
	lp->d_secpercyl = lp->d_ntracks * lp->d_nsectors;
	lp->d_ncylinders = total / lp->d_secpercyl;
	lp->d_secperunit = lp->d_ncylinders * lp->d_secpercyl;
	lp->d_partitions[RAW_PART].p_size = lp->d_secperunit;
	lp->d_checksum = 0;
	lp->d_checksum = dkcksum(lp);

	tbp = geteblk((int)lp->d_secsize);
	tbp->b_dev = dosp->dev;
	tbp->b_bcount = lp->d_secsize;
	tbp->b_blkno = lp->d_secpercyl * (lp->d_secsize / DEV_BSIZE);
	tbp->b_flags = B_BUSY | B_READ;
	tbp->b_cylin = 1;
	(*dosp->start)(tbp);
	if (biowait(tbp))
		goto done;

	buf = tbp->b_data;

	/****** XXX ******/
	if (buf[0] == 0xeb)
		error = 0;
	/****** XXX ******/

done:
	tbp->b_flags |= B_INVAL;
	brelse(tbp);
	*lp = slp;
	return error;
}

static int
scan_geo_by_tl(dosp, dvcfg, track, sector)
	struct dos_check_args *dosp;
	u_int dvcfg;
	u_int *track, *sector;
{
	struct disklabel *lp = dosp->lp;
	struct sdg_conv_element *scep;
	struct sdg_conv_table *sctp;
	int no;

	dvcfg = (DVCFG_MINOR(dvcfg) >> 12) & 0x0f;
	if (dvcfg >= sizeof(sdg_conv_table_array) / sizeof(struct sdg_conv_table *))
		return ENOENT;

	sctp = sdg_conv_table_array[dvcfg];
	if (sctp == NULL || sctp->sct_sz == 0)
		return ENOENT;

	scep = sctp->sct_pt;
	for (no = 0; no < sctp->sct_sz; no ++, scep ++)
	{
		if (autoscan_debug)
			printf("scan %d %d\n",
				(u_int32_t) scep->sce_org.sdg_tracks,
				(u_int32_t) scep->sce_org.sdg_sectors);

		if ((u_int32_t) scep->sce_org.sdg_tracks == lp->d_ntracks &&
		    (u_int32_t) scep->sce_org.sdg_sectors == lp->d_nsectors)
		{
			if (autoscan_debug)
				printf("match %d %d\n",
					(u_int32_t) scep->sce_mod.sdg_tracks,
					(u_int32_t) scep->sce_mod.sdg_sectors);
			break;
		}
	}

	if (no >= sctp->sct_sz)
		return ENOENT;

	*track = (u_int32_t) scep->sce_mod.sdg_tracks;
	*sector = (u_int32_t) scep->sce_mod.sdg_sectors;

	return 0;
}

int
scan_geo_by_ur(dosp, track, sector)
	struct dos_check_args *dosp;
	u_int *track, *sector;
{
	u_int etrack, esector;
	struct disklabel *lp = dosp->lp;
	int error;

	/* (A) original geo (55 NEC type) */
	etrack = lp->d_ntracks;
	esector = lp->d_nsectors;
	if ((error = scsi_geo_match(dosp, etrack, esector, OS_ANY)) == 0)
		goto match;

	/* (B) 92 fixed */
	etrack = 8;
	esector = 32;
	if ((error = scsi_geo_match(dosp, etrack, esector, OS_ANY)) == 0)
		goto match;

	esector = 128;
	if ((error = scsi_geo_match(dosp, etrack, esector, OS_ANY)) == 0)
		goto match;

	/* (C) round up */
	etrack = lp->d_ntracks;
	esector = round_up(lp->d_nsectors);
	if ((error = scsi_geo_match(dosp, etrack, esector, OS_ANY)) == 0)
		goto match;

match:
	if (error == 0)
	{
		*track = etrack;
		*sector = esector;
	}
	else
	{
		*track = lp->d_ntracks;
		*sector = lp->d_nsectors;
	}

	return error;
}
#endif	/* NSD_SCSIBUS > 0 */
#endif	/* NSD > 0 */

/*******************************************************************
 * IDE DISK GEOMETRY SCANNING
 *******************************************************************/
#if	NWD > 0
static u_char wdsymbol[4] = {'w', 'd', '0', 0};
#endif	/* NWD > 0 */

/*******************************************************************
 * GEOMETRY CHANGE
 *******************************************************************/
static void
change_disk_geometry(dosp)
	struct  dos_check_args *dosp;
{
	struct disklabel *lp = dosp->lp;
	struct cpu_disklabel *osdep = dosp->cdp;
	dev_t blkmjr, fixdev;

	blkmjr = major(strategy_to_blkdev(dosp->dev, dosp->start));
	fixdev = MAKEDISKDEV(blkmjr, DISKUNIT(dosp->dev), RAW_PART);

	switch (lp->d_type)
	{
	default:
		return;

#if	NSD_ATAIBUS > 0
	case DTYPE_ATAPI:
		/* XXX */
		return;
#endif	/* NSD_ATAPIBUS > 0 */

#if	NSD_SCSIBUS > 0
	case DTYPE_SCSI:
		{
		u_int total, cyls, heads, dvcfg, track, sector, busid, id, lun;
		vaddr_t addr;
		int error, unit;
		u_int8_t ebits;

		if (blkmjr != findblkmajor("sd"))
			return;

		addr = 0;
 		unit = DISKUNIT(dosp->dev);
		total = lp->d_secperunit;
		snoop_sd_dvcfg(unit, &dvcfg);
		if (autoscan_debug)
			printf("unit %x dvcfg 0x%x track %d sector %d\n",
				unit, dvcfg, lp->d_ntracks, lp->d_nsectors);

		sdsymbol[2] = unit + '0';
		if (scan_geo_by_config(sdsymbol, &track, &sector) == 0 ||
		    scan_geo_by_tl(dosp, dvcfg, &track, &sector) == 0)
		{
			lp->d_ntracks = track;
			lp->d_nsectors = sector;
			goto sdfound;
		}

		if (lookup_bootinfo_disk_geometry(fixdev, lp, osdep) == 0)
		{
			goto sdfound;
		}

		error = snoop_scsibus_info(unit, &busid, &id, &lun);
		if (error == 0 && busid == scsibios_scsibus_id &&
		    id < 7 && lun == 0)
		{
			addr = (vaddr_t)(BIOS_SCSI_PARAM + id * 4);
			ebits = lookup_bios_param(BIOS_SCSI_EQ, sizeof(char));
			ebits &= (u_int8_t) (1 << id);
		}
		else
			ebits = 0;

		if (ebits == 0)
		{
			scan_geo_by_ur(dosp, &track, &sector);
			lp->d_ntracks = track;
			lp->d_nsectors = sector;

sdfound:
			lp->d_secpercyl = lp->d_ntracks * lp->d_nsectors;
			lp->d_ncylinders = total / lp->d_secpercyl;
		}
		else
		{
			cyls = lookup_bios_param(addr + 2, sizeof(u_short));
			heads = lookup_bios_param(addr + 1, sizeof(char));

			if (cyls & 0x4000)
			{
				lp->d_ntracks = heads & 0x0f;
				lp->d_ncylinders = (cyls & 0xfff);
				lp->d_ncylinders |= ((heads & 0xf0) << 8);
			}
			else
			{
				lp->d_ntracks = heads;
				lp->d_ncylinders = (cyls & 0xfff);
			}

			lp->d_nsectors = lookup_bios_param(addr, sizeof(char));
			lp->d_secpercyl = lp->d_ntracks * lp->d_nsectors;
			if (lp->d_secpercyl == 0)
				lp->d_secpercyl = 100;
		}
		break;
		}
#endif	/* NSD > 0 */

#if	NWD > 0
	case DTYPE_ESDI:
	case DTYPE_ST506:
		{
		int unit;
		u_int total, track, sector;

		if (blkmjr != findblkmajor("wd"))
			return;

 		unit = DISKUNIT(dosp->dev);
		total = lp->d_ncylinders * lp->d_secpercyl;

		wdsymbol[2] = unit + '0';
		if (scan_geo_by_config(wdsymbol, &track, &sector) == 0)
		{
			lp->d_ntracks = track;
			lp->d_nsectors = sector;
			goto wdfound;
		}

		if (lookup_bootinfo_disk_geometry(fixdev, lp, osdep) == 0)
		{
			goto wdfound;
		}

		if (howmany(total, 8 * 17) < (1 << 16))
		{
			lp->d_ntracks =  8;
			lp->d_nsectors = 17;
		}
		else if (lp->d_subtype > 0)
		{
			lp->d_ntracks =  8 * 2;
			lp->d_nsectors = 17;
		}

wdfound:
		lp->d_secpercyl = lp->d_ntracks * lp->d_nsectors;
		lp->d_ncylinders = total / lp->d_secpercyl;
		break;
		}
#endif	/* NWD > 0 */
	}

	lp->d_secperunit = lp->d_secpercyl * lp->d_ncylinders;
	lp->d_partitions[RAW_PART].p_size = lp->d_secperunit;

	lp->d_checksum = 0;
	lp->d_checksum = dkcksum(lp);
}

/*******************************************************************
 * slice hook
 *******************************************************************/
LIST_HEAD(disk_slice_hook_tab, disk_slice_hook) disk_slice_hook_tab = 
	LIST_HEAD_INITIALIZER(&disk_slice_tab);

struct disk_slice_hook *
disk_slice_hook_scan(dev)
	dev_t dev;
{
	struct disk_slice_hook *dsp;

	for (dsp = LIST_FIRST(&disk_slice_hook_tab); dsp != NULL;
 	     dsp = LIST_NEXT(dsp, ds_chain))
	{
		if (dev == dsp->ds_dev)
			break;
	}

	return dsp;
}

struct disk_slice_hook *
disk_slice_hook_insert(dev, sector)
	dev_t dev;
	u_long sector;
{
	struct disk_slice_hook *dsp;

	dsp = disk_slice_hook_scan(dev);
	if (dsp != NULL)
	{
		dsp->ds_sector = sector;
		return dsp;
	}

	dsp = malloc(sizeof(*dsp), M_DEVBUF, M_WAITOK);
	bzero(dsp, sizeof(*dsp));
	dsp->ds_dev = dev;
	dsp->ds_sector = sector;
	LIST_INSERT_HEAD(&disk_slice_hook_tab, dsp, ds_chain);
	return dsp;
}
		
int
disk_slice_hook_remove(dsp)
	struct disk_slice_hook *dsp;
{

	LIST_REMOVE(dsp, ds_chain);
	free(dsp, M_DEVBUF);
	return 0;
}

/*******************************************************************
 * Misc
 *******************************************************************/
#include <sys/conf.h>

/* XXX:
 * Almost all drivers pass the incorrect device number
 * like MAKEDEV(0, unit, RAW_PART) or MAKEDEV(major(dev), unit, RAW_PART)
 * where dev is character device number.
 * We should fix up the device number!
 */
dev_t
strategy_to_blkdev(dev, start)
	dev_t dev;
	void (*start) __P((struct buf *));
{
	u_long i;

	for (i = 0; i < nblkdev; i ++)
	{
		if (bdevsw[i].d_strategy == start)
			break;
	}

	return makedev(i, minor(dev));
}

dev_t
findblkmajor(name)
	const char *name;
{
	struct devnametobdevmaj *nam2blk = dev_name2blk;
	int i;

	for (i = 0; nam2blk[i].d_name != NULL; i++)
		if (strncmp(name, nam2blk[i].d_name,
		    strlen(nam2blk[i].d_name)) == 0)
			return (dev_t) (nam2blk[i].d_maj);
	return (dev_t) (-1);
}

/*******************************************************************
 * Boot slice
 *******************************************************************/
#include <machine/cpu.h>
#include <machine/bootinfo.h>

static dev_t biosdev2mjr __P((dev_t));
struct biosdev2name {
	dev_t bn_dev;
	char *bn_name;
};

static struct biosdev2name biosdev2name[] = {
	{ 0x80, "wd"},
	{ 0xa0, "sd"},
	{ 0x30, "fd"},
	{ 0x90, "fd"},
	{0, NULL},
};
	
static dev_t
biosdev2mjr(dev)
	dev_t dev;
{
	struct biosdev2name *bnp;

	dev = dev & 0xf0;
	for (bnp = &biosdev2name[0]; bnp->bn_name != NULL; bnp ++)
	{
		if (bnp->bn_dev == dev)
		{
			return findblkmajor(bnp->bn_name);
		}
	}

	return (dev_t) -1;
}

int
find_bootdisk_slice(dev, strat, lp, osdep) 
	dev_t dev;
	void (*strat) __P((struct buf *));
	struct disklabel *lp;
	struct cpu_disklabel *osdep;
{
	static int found;
	struct disklabel ltmp;
	struct btinfo_bootdisk *bid;
	dev_t blkmjr;
	char *msg;
	int i;

	if (found != 0)
		return EBUSY;

	bid = lookup_bootinfo(BTINFO_BOOTDISK);
	if (bid == NULL)
		return ENOENT;

	if (bid->labelsector == -1)
		return EINVAL;

        blkmjr = biosdev2mjr(bid->biosdev);
	if (blkmjr != major(dev))
		return EINVAL;

	for (i = 0; i < osdep->cd_nslices; i ++)
	{
		if (IS_BSDFS(&osdep->cd_se[i]) == 0)
			continue;
		if (osdep->cd_se[i].se_start != bid->labelsector - 1)
			continue;

		ltmp = *lp;
		msg = read_slice_disklabel(dev, strat, &ltmp, osdep,
		   			   osdep->cd_se[i].se_start);
		if (msg != NULL)
			continue;
		if (ltmp.d_type != bid->label.type ||
		    ltmp.d_checksum != bid->label.checksum ||
		    strncmp(ltmp.d_packname, bid->label.packname, 16) != 0)
			continue;

		printf("bootdisk_slice: dev 0x%lx slno %d sector 0x%lx\n",
			(u_long) dev, i, (u_long) bid->labelsector - 1);

		disk_slice_hook_insert(
			MAKEDISKDEV(blkmjr, DISKUNIT(dev), RAW_PART),
		        bid->labelsector - 1);
		found = 1;
		return 0;
	}
	return ENOENT;
}

/*******************************************************************
 * Boot geometry hook
 *******************************************************************/
struct disk_geometry_hook {
	dev_t dh_dev;
	u_long dh_tracks;
	u_long dh_heads;
	LIST_ENTRY(disk_geometry_hook) dh_chain;
};

LIST_HEAD(disk_geometry_hook_tab, disk_geometry_hook) disk_geometry_hook_tab = 
	LIST_HEAD_INITIALIZER(&disk_geometry_hook_tab);

static int
lookup_bootinfo_disk_geometry(dev, lp, osdep)
	dev_t dev;
	struct disklabel *lp;
	struct cpu_disklabel *osdep;
{
	struct btinfo_biosgeom *big;
	struct disk_geometry_hook *dhp;
	dev_t blkmjr;
	int i;

	for (dhp = LIST_FIRST(&disk_geometry_hook_tab); dhp != NULL;
 	     dhp = LIST_NEXT(dhp, dh_chain))
	{
		if (dev == dhp->dh_dev)
		{
			lp->d_ntracks = dhp->dh_tracks;
			lp->d_nsectors = dhp->dh_heads;
			return 0;
		}
	}

	big = lookup_bootinfo(BTINFO_BIOSGEOM);
	if (big == NULL)
		return ENOENT;

	blkmjr = major(dev);
	for (i = 0; i < big->num; i++)
	{
		if (blkmjr != biosdev2mjr(big->disk[i].dev))
			continue;
		if ((big->disk[i].flags & BI_GEOM_INVALID) != 0)
		{
			if ((big->disk[i].dev & 0xf) != DISKUNIT(dev))
				continue;
			if (blkmjr != findblkmajor("wd") || 
			    DISKUNIT(dev) != 0)
				continue;
		}
		else
		{
			if (big->disk[i].cksum != osdep->cd_cksum)
				continue;
		}

		if (big->disk[i].sec == 0 || big->disk[i].head == 0)
			continue;

		dhp = malloc(sizeof(*dhp), M_DEVBUF, M_WAITOK);
		bzero(dhp, sizeof(*dhp));
		dhp->dh_dev = dev;
		dhp->dh_tracks = big->disk[i].head;
		lp->d_ntracks = big->disk[i].head;
		dhp->dh_heads = big->disk[i].sec;
		lp->d_nsectors = big->disk[i].sec;
		LIST_INSERT_HEAD(&disk_geometry_hook_tab, dhp, dh_chain);

		printf("bootinfo_disk_geometry: dev 0x%lx, tracks 0x%lx, heads 0x%lx\n",
			(u_long) dev, (u_long) lp->d_ntracks, (u_long) lp->d_nsectors);
		return 0;
	}
	return ENOENT;
}
