/*	$NecBSD: disksubr.c,v 3.36.8.3 2000/07/17 03:16:10 honda Exp $	*/
/*	$NetBSD: disksubr.c,v 1.21 1996/05/03 19:42:03 christos Exp $	*/

#define	FIXUP_OLD_DISKLABEL 	1
#define	COMPAT_OLD_DISKLABEL	1

/*
 * [NetBSD for NEC PC-98 series]
 *  Copyright (c) 1994, 1995, 1996, 1997, 1998
 *	NetBSD/pc98 porting staff. All rights reserved.
 */
/*
 * Copyright (c) 1982, 1986, 1988 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	@(#)ufs_disksubr.c	7.16 (Berkeley) 5/4/91
 */

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/buf.h>
#include <sys/malloc.h>
#include <sys/disklabel.h>
#include <sys/syslog.h>
#include <i386/i386/autoscan.h>

#define	b_cylin	b_resid

static void fixup_C_partition __P((struct disklabel *, struct cpu_disklabel *, int));
static int enter_slices_to_disklabel __P((dev_t, struct disklabel *, struct cpu_disklabel *));
static int hint_start_slice __P((dev_t, struct cpu_disklabel *));

/*
 * Attempt to read a disk label from a device
 * using the indicated stategy routine.
 * The label must be partly set up before this:
 * secpercyl, secsize and anything required for a block i/o read
 * operation in the driver's strategy/start routines
 * must be filled in before calling us.
 *
 * If dos partition table requested, attempt to load it and
 * find disklabel inside a DOS partition. Also, if bad block
 * table needed, attempt to extract it as well. Return buffer
 * for use in signalling errors if requested.
 *
 * Returns null on success and an error string on failure.
 */

static void
fixup_C_partition(lp, osdep, slno)
	struct disklabel *lp;
	struct cpu_disklabel *osdep;
	int slno;
{

	/* setup C partition, anyway */
	if (slno >= 0)
	{
		/* found BSD partition in the DOS MBR table */
		lp->d_partitions[RAW_PART - 1].p_offset =
			osdep->cd_se[slno].se_start;
		lp->d_partitions[RAW_PART - 1].p_size = 
			osdep->cd_se[slno].se_size;
	}
	else 
	{
		lp->d_partitions[RAW_PART - 1] = lp->d_partitions[RAW_PART];
	}
	
	lp->d_checksum = 0;
	lp->d_checksum = dkcksum(lp);
}

int
netbsd_start_slice(dev, osdep)
	dev_t dev;
	struct cpu_disklabel *osdep;
{
	u_long rawdev, hint;
	struct disk_slice_hook *dsp;
	int i;

	hint = DOS_CHECK_NO_HINT;
	rawdev = MAKEDISKDEV(major(dev), DISKUNIT(dev), RAW_PART);
	dsp = disk_slice_hook_scan(rawdev);
	if (dsp != NULL)
	{
		hint = dsp->ds_sector;
	}

	for (i = 0; i < osdep->cd_nslices; i ++)
	{
		if (IS_BSDFS(&osdep->cd_se[i]) == 0)
			continue;

		if (osdep->cd_se[i].se_start == 0 ||
		    osdep->cd_se[i].se_size == 0)
			continue;

		if (hint == DOS_CHECK_NO_HINT || 
		    hint == osdep->cd_se[i].se_start)
		{
			return i;
		}
	}
	return -1;
}

static int
hint_start_slice(dev, osdep)
	dev_t dev;
	struct cpu_disklabel *osdep;
{
	u_long rawdev;
	struct disk_slice_hook *dsp;
	int i;

	rawdev = MAKEDISKDEV(major(dev), DISKUNIT(dev), RAW_PART);
	dsp = disk_slice_hook_scan(rawdev);
	if (dsp == NULL)
		return -1;

	for (i = 0; i < osdep->cd_nslices; i ++)
	{
		if (dsp->ds_sector == osdep->cd_se[i].se_start)
			return i;
	}
	return -1;
}

static int
enter_slices_to_disklabel(dev, lp, osdep)
	dev_t dev;
	struct disklabel *lp;
	struct cpu_disklabel *osdep;
{
	struct disk_slice_hook *dsp;
	u_long rawdev, hint;
	int i, j, part;

    	hint = DOS_CHECK_NO_HINT;
#ifdef	COMPAT_OLD_DISKLABEL
	if (major(dev) == findblkmajor("vpd"))
#endif	/* COMPAT_OLD_DISKLABEL */
	{
		rawdev = MAKEDISKDEV(major(dev), DISKUNIT(dev), RAW_PART);
		dsp = disk_slice_hook_scan(rawdev);
		if (dsp != NULL)
		{
			hint = dsp->ds_sector;
		}
	}


	for (part = i = 0; i < osdep->cd_nslices; i++)
	{
		if (IS_BSDFS(&osdep->cd_se[i]) != 0)
			continue;
		if (hint != DOS_CHECK_NO_HINT &&
		    hint != osdep->cd_se[i].se_start)
			continue;

		/* check double entry */
		for (j = 0; j < lp->d_npartitions; j ++)
		{
			if (j == RAW_PART || j == RAW_PART - 1)
				continue;
			if (lp->d_partitions[j].p_offset ==
			    osdep->cd_se[i].se_start)
				break;
		}
		if (j < lp->d_npartitions)
			continue;

		/* find an empty slot */
		for ( ; part < MAXPARTITIONS; part ++)
			if (lp->d_partitions[part].p_fstype == FS_UNUSED &&
			    part != RAW_PART - 1 && part != RAW_PART)
				break;
		if (part == MAXPARTITIONS)
			break;

		/* register the target */
		lp->d_partitions[part].p_size = osdep->cd_se[i].se_size;
		lp->d_partitions[part].p_offset = osdep->cd_se[i].se_start;
		lp->d_partitions[part].p_fstype = osdep->cd_se[i].se_fstype;
		lp->d_partitions[part].p_frag = 0;
		lp->d_partitions[part].p_cpg = 0;
		if (part + 1 > lp->d_npartitions)
			lp->d_npartitions = part + 1;
	}

#if	0
	lp->d_checksum = 0;
	lp->d_checksum = dkcksum(lp);
#endif
	return 0;
}

char *
read_slice_disklabel(dev, strat, lp, osdep, startsector)
	dev_t dev;
	void (*strat) __P((struct buf *));
	struct disklabel *lp;
	struct cpu_disklabel *osdep;
	u_long startsector;
{
	struct disklabel *dlp;
	struct buf *bp;
	int i, error, size;
	char *msg = NULL;

	size = ((int) lp->d_secsize) < NBPG ? NBPG : lp->d_secsize;
	bp = geteblk(size);
	bp->b_dev = dev;
	bp->b_blkno = startsector + LABELSECTOR;
	bp->b_cylin = bp->b_blkno / lp->d_secpercyl;
	bp->b_blkno *= (lp->d_secsize / DEV_BSIZE);
	bp->b_bcount = lp->d_secsize;
	bp->b_flags = B_BUSY | B_READ;

	(*strat) (bp);
	error = biowait(bp);
	if (error != 0)
	{
		msg = "disk label I/O error";
		goto done;
	}

	for (i = 0; i <= lp->d_secsize - sizeof(*dlp); i += sizeof(long))
	{
	        dlp = (struct disklabel *) (bp->b_data + i);
		if (dlp->d_magic != DISKMAGIC || dlp->d_magic2 != DISKMAGIC)
		{
			if (msg == NULL)
				msg = "no disk label";
		}
		else if (dlp->d_npartitions <= MAXPARTITIONS && !dkcksum(dlp))
		{
			*lp = *dlp;
			msg = NULL;
			break;
		}
		else
			msg = "disk label corrupted";
	}

done:
	bp->b_flags |= B_INVAL;
	brelse(bp);
	return msg;
}

char *
readdisklabel(dev, strat, lp, osdep)
	dev_t dev;
	void (*strat) __P((struct buf *));
	struct disklabel *lp;
	struct cpu_disklabel *osdep;
{
	struct dkbad *bdp;
	struct buf *bp;
	char *msg = NULL;
	int error, i, size, slno, hslno;
	struct dos_check_args dos;
	u_long startsector;
	int osdep_alloc = 0;
	struct dkbad *db;
	daddr_t lastsec;
	dev_t fixdev;

	/*****************************************************
 	 * osdep allocation
	 *****************************************************/
	if (osdep == NULL)
	{
		osdep_alloc = 1;
		osdep = malloc(sizeof(*osdep), M_TEMP, M_WAITOK);
		bzero(osdep, sizeof(*osdep));
	}

	/*****************************************************
 	 * get blk device num
	 *****************************************************/
	fixdev = strategy_to_blkdev(dev, strat);

	/*****************************************************
 	 * setup default disklabel
	 *****************************************************/
	if (lp->d_secsize == 0)
		lp->d_secsize = DEV_BSIZE;
	if (lp->d_secperunit == 0)
		lp->d_secperunit = 0x1fffffff;
	if (lp->d_secpercyl == 0)
		lp->d_secpercyl = 100;

	for (i = 0; i < RAW_PART; i++)
	{
		lp->d_partitions[i].p_fstype = FS_UNUSED;
		lp->d_partitions[i].p_size = 0;
		lp->d_partitions[i].p_offset = 0;
	}
	if (lp->d_partitions[i].p_size == 0)
		lp->d_partitions[i].p_size = 0x1fffffff;
	lp->d_partitions[i].p_offset = 0;
	lp->d_npartitions = RAW_PART + 1;

	/*****************************************************
 	 * setup cpudisklabel
	 *****************************************************/
	memset(&dos, 0, sizeof(dos));
	dos.lp = lp;
	dos.cdp = osdep;
	dos.start = strat;
	dos.dev = dev;
	dos.flag |= GEO_CHG_EN;

	error = read_cpudisklabel(&dos);
	if (error != 0)
	{
		hslno = slno = -1;
	}
	else
	{
		find_bootdisk_slice(fixdev, strat, lp, osdep);
		slno = netbsd_start_slice(fixdev, osdep);
		if (slno >= 0)
			hslno = slno;
		else
			hslno = hint_start_slice(fixdev, osdep);
	}
	fixup_C_partition(lp, osdep, hslno);

	if (slno < 0)
		startsector = 0;
	else
		startsector = osdep->cd_se[slno].se_start;

	/*****************************************************
 	 * obtain a BSD disklabel
	 *****************************************************/
	msg = read_slice_disklabel(fixdev, strat, lp, osdep, startsector);
	enter_slices_to_disklabel(fixdev, lp, osdep);
	if (msg != NULL)
		goto done;

#ifdef	FIXUP_OLD_DISKLABEL
	if (lp->d_partitions[RAW_PART - 1].p_size == 0) 
	{
		fixup_C_partition(lp, osdep, hslno);
	}
#endif	/* FIXUP_OLD_DISKLABEL */

	/*****************************************************
 	 * obtain BAD144 data 
	 *****************************************************/
#define DKBAD_MAGIC 0x4321

	if ((lp->d_flags & D_BADSECT) == 0)
		goto done;

	bdp = &osdep->bad;
	if (lp->d_partitions[0].p_offset > 0)
		lastsec = lp->d_partitions[2].p_offset +
			  lp->d_partitions[2].p_size;
	else
		lastsec = lp->d_secperunit;

	size = ((int) lp->d_secsize) < NBPG ? NBPG : lp->d_secsize;
	bp = geteblk(size);
	bp->b_dev = dev;
	for (i = 0; i < 10 && i < lp->d_nsectors; i += 2)
	{
		bp->b_flags = B_BUSY | B_READ;
		bp->b_blkno = lastsec - lp->d_nsectors + i;
		bp->b_cylin = bp->b_blkno / lp->d_secpercyl;
		bp->b_blkno *= (lp->d_secsize / DEV_BSIZE);
		bp->b_bcount = lp->d_secsize;

		(*strat) (bp);
		error = biowait(bp);
		if (error != 0)
		{
			msg = "bad sector table I/O error";
			break;
		}

		db = (struct dkbad *)(bp->b_data);
		if (!db->bt_mbz && db->bt_flag == DKBAD_MAGIC)
		{
			msg = NULL;
			*bdp = *db;
			break;
		}
		else
			msg = "bad sector table corrupted";
	} 
	bp->b_flags |= B_INVAL;
	brelse(bp);

done:
	if (osdep_alloc != 0)
	{
		free(osdep, M_TEMP);
	}
	return (msg);
}

/*
 * Check new disk label for sensibility
 * before setting it.
 */
int
setdisklabel(olp, nlp, openmask, osdep)
	struct disklabel *olp, *nlp;
	u_long openmask;
	struct cpu_disklabel *osdep;
{
	int i;
	struct partition *opp, *npp;

	/* sanity clause */
	if (nlp->d_secpercyl == 0 || nlp->d_secsize == 0
		|| (nlp->d_secsize % DEV_BSIZE) != 0)
			return(EINVAL);

	/* special case to allow disklabel to be invalidated */
	if (nlp->d_magic == 0xffffffff) {
		*olp = *nlp;
		return (0);
	}

	if (nlp->d_magic != DISKMAGIC || nlp->d_magic2 != DISKMAGIC ||
	    dkcksum(nlp) != 0)
		return (EINVAL);

	/* XXX missing check if other dos partitions will be overwritten */

	while (openmask != 0) {
		i = ffs(openmask) - 1;
		openmask &= ~(1 << i);
		if (nlp->d_npartitions <= i)
			return (EBUSY);
		opp = &olp->d_partitions[i];
		npp = &nlp->d_partitions[i];
		if (npp->p_offset != opp->p_offset || npp->p_size < opp->p_size)
			return (EBUSY);
		/*
		 * Copy internally-set partition information
		 * if new label doesn't include it.		XXX
		 */
		if (npp->p_fstype == FS_UNUSED && opp->p_fstype != FS_UNUSED)
		{
			npp->p_fstype = opp->p_fstype;
			npp->p_fsize = opp->p_fsize;
			npp->p_frag = opp->p_frag;
			npp->p_cpg = opp->p_cpg;
		}
	}
 	nlp->d_checksum = 0;
 	nlp->d_checksum = dkcksum(nlp);
	*olp = *nlp;
	return (0);
}


/*
 * Write disk label back to device after modification.
 */
int
writedisklabel(dev, strat, lp, osdep)
	dev_t dev;
	void (*strat) __P((struct buf *));
	struct disklabel *lp;
	struct cpu_disklabel *osdep;
{
	struct buf *bp;
	struct disklabel *dlp;
	int error, size, i, slno;
	struct dos_check_args dos;
	int osdep_alloc = 0;
	u_long startsector;
	dev_t fixdev;

	/*****************************************************
 	 * osdep allocation
	 *****************************************************/
	if (osdep == NULL)
	{
		osdep_alloc = 1;
		osdep = malloc(sizeof(*osdep), M_TEMP, M_WAITOK);
		bzero(osdep, sizeof(*osdep));
	}

	/*****************************************************
 	 * get block device number
	 *****************************************************/
	fixdev = strategy_to_blkdev(dev, strat);

	/*****************************************************
 	 * check dos partition.
	 *****************************************************/
	if (lp->d_partitions[RAW_PART - 1].p_size == 0)
	{
		error = EINVAL;
		goto out;
	}

	memset(&dos, 0, sizeof(dos));
	dos.lp = lp;
	dos.cdp = osdep;
	dos.start = strat;
	dos.dev = dev;

	error = read_cpudisklabel(&dos);
	if (error == EIO)
		goto out;
	if (error != 0)
	{
		startsector = 0;
	}
	else
	{
		slno = netbsd_start_slice(fixdev, osdep);
		if (slno < 0)
		{
			error = EINVAL;
			goto out;
		}
		startsector = osdep->cd_se[slno].se_start;
	}

	/* XXX:
	 * Final consistency check.
	 */
	
	if (startsector != lp->d_partitions[RAW_PART - 1].p_offset)
	{
		error = EINVAL;
		goto out;
	}
	
	/*****************************************************
 	 * find a BSD partition and write it
	 *****************************************************/
	size = ((int) lp->d_secsize) < NBPG ? NBPG : lp->d_secsize;
	bp = geteblk(size);
	bp->b_dev = dev;
	bp->b_blkno = startsector + LABELSECTOR;
	bp->b_cylin = bp->b_blkno / lp->d_secpercyl;
	bp->b_blkno *= (lp->d_secsize / DEV_BSIZE);
	bp->b_bcount = lp->d_secsize;
	bp->b_flags = B_BUSY | B_READ;

	(*strat) (bp);
	error = biowait(bp);
	if (error != 0)
		goto done;

	for (i = 0; i <= lp->d_secsize - sizeof(*dlp); i += sizeof(long))
	{
	        dlp = (struct disklabel *) (bp->b_data + i);
		if (dlp->d_magic == DISKMAGIC && dlp->d_magic2 == DISKMAGIC &&
		    dkcksum(dlp) == 0)
		{
			*dlp = *lp;
			bp->b_flags = B_BUSY | B_WRITE;
			(*strat) (bp);
			error = biowait(bp);
			goto done;
		}
	}
	error = ESRCH;

done:
	bp->b_flags |= B_INVAL;
	brelse(bp);

out:
	if (osdep_alloc != 0)
	{
		free(osdep, M_TEMP);
	}
	return (error);
}

/*
 * Determine the size of the transfer, and make sure it is
 * within the boundaries of the partition. Adjust transfer
 * if needed, and signal errors or early completion.
 */
int
bounds_check_with_label(bp, lp, wlabel)
	struct buf *bp;
	struct disklabel *lp;
	int wlabel;
{
	struct partition *p = lp->d_partitions + DISKPART(bp->b_dev);
	daddr_t labelsector, sz, psize, poffs;
	daddr_t sc;

	sz = howmany(bp->b_bcount, DEV_BSIZE);
	labelsector = lp->d_partitions[RAW_PART - 1].p_offset + LABELSECTOR;

	if (lp->d_secsize == DEV_BSIZE)
	{
		sc = 1;
		psize = p->p_size;
		poffs = p->p_offset;
	}
	else
	{
		/* XXX:
		 * if secsize != DEV_BSIZE, should convert all partition addrs
		 * into values based on abstract device block size.
		 */
		sc = lp->d_secsize >> DEV_BSHIFT;
		if (sc < 1)
		{
			bp->b_error = EINVAL;
			goto bad;
		}
		psize = ((daddr_t) p->p_size) * sc;
		poffs = ((daddr_t) p->p_offset) * sc;
		labelsector *= sc;
	}

	/* Assumption:
	 * 1) bp->b_blkno < 0 already checked by a upper layer func().
	 * 2) bp->b_bcount alreay limited by minphys()!
	 */
	if (((u_long) bp->b_blkno) + ((u_long) sz) > ((u_long) psize))
	{
		sz = psize - bp->b_blkno;
		if (sz == 0)
		{
			bp->b_resid = bp->b_bcount;
			goto done;
		}

		if (sz < 0)
		{
			bp->b_error = EINVAL;
			goto bad;
		}

		bp->b_bcount = sz << DEV_BSHIFT;
	}

	poffs = bp->b_blkno + poffs;
	/* A remaining case to be checked. */
	if (poffs + sz < 0)
	{
		bp->b_error = EINVAL;
		goto bad;
	}

	if (poffs <= labelsector &&
#if LABELSECTOR != 0
	    poffs + sz > labelsector &&
#endif
	    (bp->b_flags & B_READ) == 0 && !wlabel)
	{
		bp->b_error = EROFS;
		goto bad;
	}

	bp->b_cylin = poffs / sc / lp->d_secpercyl;
	return (1);

bad:
	bp->b_flags |= B_ERROR;
done:
	return (0);
}
