/*	$NecBSD: vpd.c,v 3.20.4.3 2000/07/18 05:48:50 honda Exp $	*/

#define	VPD_SAFE_SERVER
#define	VPD_PROTECT_OTHERS
/*
 * [NetBSD for NEC PC-98 series]
 *  Copyright (c) 1997, 1998, 2000
 *	 NetBSD/pc98 porting staff. All rights reserved.
 *  Copyright (c) 1997, 1998, 2000
 *	Naofumi HONDA. All rights reserved.
 */
/*
 * Copyright (c) 1988 University of Utah.
 * Copyright (c) 1990, 1993
 *	The Regents of the University of California.  All rights reserved.
 *
 * This code is derived from software contributed to Berkeley by
 * the Systems Programming Group of the University of Utah Computer
 * Science Department.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * from: Utah $Hdr: vn.c 1.13 94/04/02$
 *
 *	@(#)vn.c	8.6 (Berkeley) 4/1/94
 */

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/namei.h>
#include <sys/proc.h>
#include <sys/errno.h>
#include <sys/buf.h>
#include <sys/malloc.h>
#include <sys/ioctl.h>
#include <sys/disklabel.h>
#include <sys/device.h>
#include <sys/disk.h>
#include <sys/stat.h>
#include <sys/mount.h>
#include <sys/vnode.h>
#include <sys/file.h>
#include <sys/uio.h>
#include <sys/conf.h>
#include <sys/lock.h>
#include <sys/kthread.h>
#include <sys/signalvar.h>
#include <vm/vm.h>

#include <miscfs/specfs/specdev.h>

#include <dev/isa/ccbque.h>
#include <machine/vpdio.h>
#include <i386/i386/autoscan.h>

#define	VPD_NMAXIO	64
#define	VPDDRV_NDRV	0x000ff
#define	VPDDRV_BOOTCONF	0x10000

#define	vpdunit(x)	((minor(x)) / MAXPARTITIONS)
#define	vpdpart(x)	((x) % MAXPARTITIONS)
#define	vpdmajor(x)	major((x))

struct vpdbuf {
	struct buf	vb_buf;		/* strategy buf */
	struct buf	*vb_obp;	/* shadowed buffer */

	TAILQ_ENTRY(vpdbuf) vb_chain;	/* free buffer chain */
};

struct vpd_lower_vnode {
	dev_t vu_tdev;
	struct vnode *vu_vp;

	struct disklabel vu_d;
	struct cpu_disklabel vu_cd;

	int vu_openf;
	int vu_refc;

	LIST_ENTRY(vpd_lower_vnode) vu_chain;
};

struct vpdsrv_softc {
	struct device	sc_dev;		/* device sosftc */

	struct vpd_softc **sc_vpd;
	int sc_nvpd;
	int sc_vpdsslp;
};

struct vpd_softc {
	struct device	sc_dev;		/* device sosftc */
	struct vpdsrv_softc *sc_dsc;

	char sc_xname[16];
	dev_t sc_tdev;
	int   sc_slno;
	struct vpd_lower_vnode *sc_vup;

	struct vpd_slice {
#define	VPD_SLICE_LOADED	0x0001
#define	VPD_SLICE_WLABEL	0x0002
#define	VPD_SLICE_LABELLING	0x0004
#define	VPD_SLICE_ACTIVE	0x0008
#define	VPD_SLICE_BLOCKED	0x0010
		struct disk	sl_dkdev;	/* generic disk device info */
		daddr_t		sl_sblk;
		daddr_t		sl_plblk;
		daddr_t		sl_phblk;
		u_int		sl_flags;	/* flags */
		char		sl_xname[16];
		struct disk_slice_hook *sl_dsp;
	} sc_sl;

	struct lock	sc_lock;

#define	VPD_ACTIVE	0x00001
#define	VPD_RDONLY	0x10000
#define	VPD_SERVER	0x20000
	u_int		sc_flags;	/* flags */
#define	VPD_STATE_ZERO		0
#define	VPD_STATE_CTL		1
#define	VPD_STATE_OPEN		2
#define	VPD_STATE_CLOSING	3
	int		sc_state;

	int		sc_nbp;		/* dummy count */
	int		sc_nbuf;	/* num of the current io */
	int		sc_maxnbuf;	/* max buffers */
	struct buf	sc_btab;	/* buffers tab */
};

LIST_HEAD(vu_tab, vpd_lower_vnode) vu_tab = LIST_HEAD_INITIALIZER(&vu_tab);
static struct vpdsrv_softc *vpdsrv_boot;

static void vpdserverstart __P((void *));
static void vpdiodone __P((struct buf *));
static int vpdgetdisklabel __P((struct vpd_softc *));
static void vpdsafeserver __P((void *));
static int vpdisactive __P((struct vpd_softc *));
static void vpdio __P((struct vpd_softc *, int));
static int vpdvnodeopen __P((struct vpd_softc *, dev_t, int));
static int vpdvnodeclose __P((struct vpd_softc *));
static dev_t vpdgetbdev __P((char *, int));
static int vpdgetdefaultlabel __P((struct vpd_softc *, struct disklabel *));

cdev_decl(vpd);
bdev_decl(vpd);

#define	VPD_SOFTC(unit)	(vpd_cd.cd_devs[(unit)])
#define	VPD_NDEVS	(vpd_cd.cd_ndevs)
#define	b_cylin	b_resid

#define	vpdlock(sc) lockmgr(&(sc)->sc_lock, LK_EXCLUSIVE, NULL)
#define	vpdunlock(sc) lockmgr(&(sc)->sc_lock, LK_RELEASE, NULL)

GENERIC_CCB_ASSERT(vpd, vpdbuf)
GENERIC_CCB_STATIC_ALLOC(vpd, vpdbuf)
GENERIC_CCB(vpd, vpdbuf, vb_chain)

static int vpdmatch __P((struct device *, struct cfdata *, void *));
static void vpdattach __P((struct device *, struct device *, void *));
static int vpdsrvmatch __P((struct device *, struct cfdata *, void *));
static void vpdsrvattach __P((struct device *, struct device *, void *));

extern struct cfdriver vpdsrv_cd;
struct cfattach vpdsrv_ca = {
	sizeof(struct vpdsrv_softc), vpdsrvmatch, vpdsrvattach
};

extern struct cfdriver vpd_cd;
struct cfattach vpd_ca = {
	sizeof(struct vpd_softc), vpdmatch, vpdattach
};

static int
vpdsrvmatch(parent, match, aux)
	struct device *parent;
	struct cfdata *match;
	void *aux;
{

	return 1;
}

static void
vpdsrvattach(parent, self, aux)
	struct device *parent, *self;
	void *aux;
{
	struct cfdata *cf = self->dv_cfdata;
	struct vpdsrv_softc *dsc = (void *) self;
	struct device *m;
	u_long flags;
	int i, units;

	flags = cf->cf_flags;

	units = (flags & VPDDRV_NDRV);
	if (units == 0)
		units = 4;
	if ((flags & VPDDRV_BOOTCONF) != 0 && vpdsrv_boot == NULL)
		vpdsrv_boot = dsc;
	printf("\n");

	dsc->sc_vpd = malloc(sizeof(struct vpd_softc *) * units,
			     M_DEVBUF, M_WAITOK);
	for (i = 0; i < units; i ++)
	{
		m = config_found(self, NULL, NULL);
		if (m == NULL)
			continue;
		dsc->sc_vpd[dsc->sc_nvpd ++] = (void *) m;
	}

	kthread_create_deferred(vpdserverstart, dsc);
}

static int
vpdmatch(parent, match, aux)
	struct device *parent;
	struct cfdata *match;
	void *aux;
{

	return 1;
}

static void
vpdattach(parent, self, aux)
	struct device *parent, *self;
	void *aux;
{
	struct vpd_softc *sc = (void *) self;
	struct vpd_slice *slp;

	printf("\n");
	sc->sc_dsc = (void *) parent;
	slp = &sc->sc_sl;
	strncpy(slp->sl_xname, sc->sc_dev.dv_xname, sizeof(slp->sl_xname));
	slp->sl_dkdev.dk_name = slp->sl_xname;
	lockinit(&sc->sc_lock, PLOCK | PCATCH, "vpdlk", 0, 0);
}

int
vpdopen(dev, flags, mode, p)
	dev_t dev;
	int flags, mode;
	struct proc *p;
{
	struct vpd_softc *sc;
	struct vpd_slice *slp;
	int error, unit, part, pmask;

	unit = vpdunit(dev);
	if (unit >= VPD_NDEVS)
		return ENXIO;
	if ((sc = VPD_SOFTC(unit)) == NULL)
		return ENXIO;

	if ((error = vpdlock(sc)) != 0)
		return error;

	part = vpdpart(dev);
	pmask = (1 << part);
	slp = &sc->sc_sl;

	switch (sc->sc_state)
	{
	default:
	case VPD_STATE_CTL:
		error = EBUSY;
		goto bad;

	case VPD_STATE_ZERO:
		if (part == VPD_SLICE_CTL)
		{
			sc->sc_state = VPD_STATE_CTL;
			goto bad;
		}

		error = vpdvnodeopen(sc, sc->sc_tdev, sc->sc_slno);
		if (error != 0)
			goto bad;
		/* go through */
	
	case VPD_STATE_OPEN:
		if ((slp->sl_flags & VPD_SLICE_ACTIVE) == 0)
		{
			error = ENXIO;
			goto bad;
		}

		if ((slp->sl_flags & VPD_SLICE_BLOCKED) != 0)
		{
			if ((part != RAW_PART && part != RAW_PART - 1) ||
			    (flags & FWRITE) != 0)
			{
				error = EBUSY;
				goto bad;
			}
		}
		break;
	}

	if ((flags & FWRITE) && (sc->sc_flags & VPD_RDONLY))
	{
		error = EROFS;
		goto bad;
	}

	if (slp->sl_dkdev.dk_openmask == 0)
	{
		if ((slp->sl_flags & VPD_SLICE_LOADED) == 0)
		{
			vpdgetdisklabel(sc);
			slp->sl_flags |= VPD_SLICE_LOADED;
		}
	}
	else
	{
		if ((slp->sl_flags & VPD_SLICE_LOADED) == 0)
		{
			error = EIO;
			goto bad;
		}
	}

	if (part != RAW_PART &&
	    (part >= slp->sl_dkdev.dk_label->d_npartitions ||
	     slp->sl_dkdev.dk_label->d_partitions[part].p_fstype == FS_UNUSED))
	{
		error = ENXIO;
		goto bad;
	}

	switch (mode)
	{
	case S_IFCHR:
		slp->sl_dkdev.dk_copenmask |= pmask;
		break;

	case S_IFBLK:
		slp->sl_dkdev.dk_bopenmask |= pmask;
		break;
	}
	slp->sl_dkdev.dk_openmask =
	    slp->sl_dkdev.dk_copenmask | slp->sl_dkdev.dk_bopenmask;
	vpdunlock(sc);
	return error;

bad:
	if (sc->sc_state == VPD_STATE_OPEN && vpdisactive(sc) == 0)
	{
		vpdvnodeclose(sc);
	}
	vpdunlock(sc);
	return error;
}

int
vpdclose(dev, flags, mode, p)
	dev_t dev;
	int flags, mode;
	struct proc *p;
{
	int unit = vpdunit(dev);
	struct vpd_slice *slp;
	struct vpd_softc *sc = VPD_SOFTC(unit);
	int error, part;

	if ((error = vpdlock(sc)) != 0)
		return (error);

	part = vpdpart(dev);
	slp = &sc->sc_sl;

	switch (sc->sc_state)
	{
	case VPD_STATE_CTL:
		sc->sc_state = VPD_STATE_ZERO;
		break;

	case VPD_STATE_ZERO:
		printf("%s: already closed\n", sc->sc_dev.dv_xname);
		break;

	case VPD_STATE_OPEN:
		switch (mode)
		{
		case S_IFCHR:
			slp->sl_dkdev.dk_copenmask &= ~(1 << part);
			break;

		case S_IFBLK:
			slp->sl_dkdev.dk_bopenmask &= ~(1 << part);
			break;
		}

		slp->sl_dkdev.dk_openmask =
		    slp->sl_dkdev.dk_copenmask | slp->sl_dkdev.dk_bopenmask;

		if (slp->sl_dkdev.dk_openmask == 0)
		{
			slp->sl_flags &= ~VPD_SLICE_LOADED;
		}

		if (vpdisactive(sc) == 0)
		{
			vpdvnodeclose(sc);
		}
		break;

	case VPD_STATE_CLOSING:
		printf("%s: closing state\n", sc->sc_dev.dv_xname);
		break;
	}

	vpdunlock(sc);
	return (0);
}

void
vpdstrategy(bp)
	register struct buf *bp;
{
	int unit = vpdunit(bp->b_dev);
	struct vpd_slice *slp;
	struct vpd_softc *sc = VPD_SOFTC(unit);
	struct vpdsrv_softc *dsc = sc->sc_dsc;
	struct disklabel *lp;
	int part, s;

	slp = &sc->sc_sl;
	if (sc->sc_state != VPD_STATE_OPEN)
	{
		bp->b_error = ENXIO;
		goto bad;
	}

	if (bp->b_bcount == 0)
		goto done;

	lp = slp->sl_dkdev.dk_label;
	if (bp->b_blkno < 0 || (bp->b_bcount % lp->d_secsize) != 0)
	{
		bp->b_error = EINVAL;
		goto bad;
	}

	part = vpdpart(bp->b_dev);
	if (part != RAW_PART && 
	    bounds_check_with_label(bp, lp, 
	     (slp->sl_flags & (VPD_SLICE_LABELLING | VPD_SLICE_WLABEL))) <= 0)
		goto done;
	
#ifdef	VPD_PROTECT_OTHERS
	if ((bp->b_flags & B_READ) == 0)
	{
		daddr_t bn;

		bn = lp->d_partitions[part].p_offset * 
		     (lp->d_secsize / dbtob(1)) + bp->b_blkno;
		if (bn < sc->sc_sl.sl_plblk || bn >= sc->sc_sl.sl_phblk)
		{
			if (((slp->sl_flags & VPD_SLICE_WLABEL) != 0 && 
			     bn < sc->sc_sl.sl_sblk) == 0)
			{
				bp->b_error = EROFS;
				goto bad;
			}
		}
	}
#endif	/* VPD_PROTECT_OTHERS */
		
	s = splbio();
	disksort(&sc->sc_btab, bp);
	if ((sc->sc_flags & VPD_SERVER) != 0)
	{
		if (dsc->sc_vpdsslp > 0)
		{
			dsc->sc_vpdsslp = 0;
			wakeup(&dsc->sc_vpdsslp);
		}
	}
	else
	{
		if ((sc->sc_flags & VPD_ACTIVE) == 0)
			vpdio(sc, 1);
	}
	splx(s);
	return;

bad:
	bp->b_flags |= B_ERROR;
done:
	bp->b_resid = bp->b_bcount;
	biodone(bp);
	return;
}

void
vpdio(sc, nowait)
	struct vpd_softc *sc;
	int nowait;
{
	struct buf *bp;
	struct vpd_slice *slp;
	struct vpdbuf *nbp;
	daddr_t bn;
	struct disklabel *lp;

	sc->sc_flags |= VPD_ACTIVE;

restart:
	bp = sc->sc_btab.b_actf;
	if (bp == NULL)
	{
		sc->sc_flags &= ~VPD_ACTIVE;
		return;
	}

	sc->sc_btab.b_actf = bp->b_actf;
	slp = &sc->sc_sl;
	lp = slp->sl_dkdev.dk_label;

	nbp = vpd_get_ccb(nowait);			/* sleep OK */
	if (nbp == NULL)
	{
		sc->sc_flags &= ~VPD_ACTIVE;
		return;
	}

	sc->sc_nbuf ++;
	if ((bp->b_flags & B_READ) == 0)
		sc->sc_vup->vu_vp->v_numoutput++;

	bn = bp->b_blkno;
	if (vpdpart(bp->b_dev) != RAW_PART)
	{
		register struct partition *p;

		p = lp->d_partitions + vpdpart(bp->b_dev); 
		bn += p->p_offset * (lp->d_secsize / dbtob(1));
	}

	nbp->vb_buf.b_flags = bp->b_flags | B_CALL | B_NOCACHE;
	nbp->vb_buf.b_bcount = bp->b_bcount;
	nbp->vb_buf.b_bufsize = bp->b_bufsize;
	nbp->vb_buf.b_error = 0;
	nbp->vb_buf.b_data = bp->b_data;
	nbp->vb_buf.b_blkno = bn;
	nbp->vb_buf.b_proc = bp->b_proc;
	nbp->vb_buf.b_iodone = vpdiodone;

	nbp->vb_buf.b_resid = 0;		/* trick XXX */

	nbp->vb_buf.b_validoff = 0;
	nbp->vb_buf.b_validend = 0;
	nbp->vb_buf.b_dirtyoff = 0;
	nbp->vb_buf.b_dirtyend = bp->b_bcount;

#ifdef	__BROKEN_VPD_AND_VNODE_STRATEGY
	/* XXX:
	 * The buffer entered into the clean buffer chain of the
	 * vnode. It's too dangerous becuase brelvp() of vpdiodone()
	 * will be called in interrupt context (vs. chain chasing 
	 * of vinvalbuf()).
	 */
	nbp->vb_buf.b_vnbufs.le_next = NOLIST;
	nbp->vb_buf.b_vp = NULLVP;
	bgetvp(sc->sc_vup->vu_vp, &nbp->vb_buf);
#else	/* !__BROKEN */
	/* do the same things as bgetvp()!  */
	nbp->vb_buf.b_vnbufs.le_next = NOLIST;
	nbp->vb_buf.b_vp = sc->sc_vup->vu_vp;
	nbp->vb_buf.b_dev = sc->sc_vup->vu_tdev;
	VHOLD(sc->sc_vup->vu_vp);
#endif	/* __BORKEN */

	nbp->vb_obp = bp;

	VOP_STRATEGY(&nbp->vb_buf);
	goto restart;
}

static void
vpdserverstart(arg)
	void *arg;
{
	struct vpdsrv_softc *dsc = arg;
	struct vpd_softc *sc;
	int i;

	for (i = 0; i < dsc->sc_nvpd; i ++)
	{
		sc = dsc->sc_vpd[i];
		sc->sc_flags |= VPD_SERVER;
	}

	(void) kthread_create(vpdsafeserver, dsc, NULL, "vpd_kthread");
	printf("%s: thread server starting (%d)\n",
		dsc->sc_dev.dv_xname, dsc->sc_nvpd);
}

void
vpdsafeserver(arg)
	void *arg;
{
	struct vpdsrv_softc *dsc = arg;
	struct vpd_softc *sc;
	struct proc *p;
	int error, s, i;

	p = curproc;				/* XXX */
	while (1)
	{
		for (i = 0; i < dsc->sc_nvpd; i ++)
		{
			sc = dsc->sc_vpd[i];
		    	if (sc->sc_state != VPD_STATE_OPEN)
				continue;
			if ((sc->sc_flags & VPD_SERVER) == 0)
				continue;
			if (sc->sc_btab.b_actf == NULL)
				continue;

			s = splbio();
			if ((sc->sc_flags & VPD_ACTIVE) == 0)
			{
				vpdio(sc, 0);
			}
			splx(s);
		}

		dsc->sc_vpdsslp = 1;
		error = tsleep((caddr_t) &dsc->sc_vpdsslp,
				PRIBIO, "vpdsrv", 0);
		if (error != 0)
			CLRSIG(p, CURSIG(p));
	}
}

void
vpdiodone(tbp)
	struct buf *tbp;
{
	struct vpdbuf *vbp = (struct vpdbuf *) tbp;
	struct buf *bp = vbp->vb_obp;
	struct vpd_softc *sc = VPD_SOFTC(vpdunit(bp->b_dev));
	int s;

	if (tbp->b_flags & B_ERROR)
	{
		bp->b_flags |= B_ERROR;
		bp->b_error = biowait(tbp);
	}

	bp->b_resid = tbp->b_resid;

	s = splbio();
	if (tbp->b_vp != NULL)
		brelvp(tbp);
	vpd_free_ccb(vbp);
	sc->sc_nbuf --;
	if ((sc->sc_flags & (VPD_SERVER | VPD_ACTIVE)) == 0)
		vpdio(sc, 1);
	splx(s);

	biodone(bp);
	if (sc->sc_state == VPD_STATE_CLOSING)
	{
		wakeup(&sc->sc_state);
	}
}

int
vpdread(dev, uio, flags)
	dev_t dev;
	struct uio *uio;
	int flags;
{
	int unit = vpdunit(dev);
	struct vpd_softc *sc = VPD_SOFTC(unit);

	if (sc->sc_state != VPD_STATE_OPEN)
		return (ENXIO);

	return (physio(vpdstrategy, NULL, dev, B_READ, minphys, uio));
}

int
vpdwrite(dev, uio, flags)
	dev_t dev;
	struct uio *uio;
	int flags;
{
	int unit = vpdunit(dev);
	struct vpd_softc *sc = VPD_SOFTC(unit);

	if (sc->sc_state != VPD_STATE_OPEN)
		return (ENXIO);

	return (physio(vpdstrategy, NULL, dev, B_WRITE, minphys, uio));
}

static int
vpdisactive(sc)
	struct vpd_softc *sc;
{
	struct vpd_slice *slp; 

	slp = &sc->sc_sl;
	if (slp->sl_dkdev.dk_openmask != 0)
	{
		return EBUSY;
	}
	return 0;
}

int
vpdioctl(dev, cmd, data, flag, p)
	dev_t dev;
	u_long cmd;
	caddr_t data;
	int flag;
	struct proc *p;
{
	int unit = vpdunit(dev);
	int i, error, slno, part;
	struct vpd_slice *slp; 
	struct vpd_softc *sc = VPD_SOFTC(unit);
	struct vpdsrv_softc *dsc = sc->sc_dsc;
	struct vpd_ioctl *vio;
	dev_t tdev;

	error = 0;
	part = vpdpart(dev);
	slp = &sc->sc_sl;
	vio = (struct vpd_ioctl *) data;

	switch (cmd)
	{
	case VPDIOCSET:
		error = suser(p->p_ucred, &p->p_acflag);
		if (error)
			return error;
		if ((error = vpdlock(sc)) != 0)
			return error;
		if (sc->sc_state != VPD_STATE_CTL)
		{
			error = EBUSY;
			goto out;
		}

		if (vio->vpd_xname[0] == 0)
		{
			sc->sc_tdev = 0;
			goto out;
		}

		vio->vpd_xname[15] = 0;
		slno = vio->vpd_slno;
		tdev = vpdgetbdev(vio->vpd_xname, vio->vpd_unit);
		if (tdev == 0)
		{
			error = EINVAL;
			goto out;
		}

		if (sc->sc_tdev == tdev && sc->sc_slno == slno)
			goto out;

		for (i = 0; i < dsc->sc_nvpd; i ++)
		{
			struct vpd_softc *tsc;

			tsc = dsc->sc_vpd[i];
			if (tsc->sc_tdev == tdev && tsc->sc_slno == slno)
			{
				error = EBUSY;
				goto out;
			}
		}

		error = vpdvnodeopen(sc, tdev, slno);
		if (error != 0)
			goto out;

		vpdvnodeclose(sc);
		strncpy(sc->sc_xname, vio->vpd_xname, sizeof(sc->sc_xname));
		printf("%s: connecting %s%ds%d\n",
			sc->sc_dev.dv_xname, vio->vpd_xname, vio->vpd_unit,
			sc->sc_slno);

out:
		sc->sc_state = VPD_STATE_CTL;
		vpdunlock(sc);
		return error;

	default:
		break;
	}

	if (sc->sc_state != VPD_STATE_OPEN)
		return ENXIO;
	
	switch (cmd) 
	{
	case VPDIOGINFO:
	{
		struct vpd_unit_info *vip = (struct vpd_unit_info *) data;

		strncpy(vip->vpd_xname, sc->sc_xname, sizeof(sc->sc_xname));
		vip->vpd_unit = DISKUNIT(sc->sc_tdev);
		vip->vpd_slno = sc->sc_slno;
		vip->vpd_d = sc->sc_vup->vu_d;
		vip->vpd_cd = sc->sc_vup->vu_cd;
		return 0;
	}
		
	case DIOCGDINFO:
		*(struct disklabel *) data = *(slp->sl_dkdev.dk_label);
		return 0;
	
	case DIOCGPART:
		((struct partinfo *) data)->disklab = slp->sl_dkdev.dk_label;
		((struct partinfo *) data)->part =
			&slp->sl_dkdev.dk_label->d_partitions[vpdpart(dev)];
		return 0;
	
	case DIOCWDINFO:
	case DIOCSDINFO:
		error = suser(p->p_ucred, &p->p_acflag);
		if (error)
			return (error);
		if ((flag & FWRITE) == 0)
			return EBADF;

		if ((error = vpdlock(sc)) != 0)
			return error;

		slp->sl_flags |= VPD_SLICE_LABELLING;
		error = setdisklabel(slp->sl_dkdev.dk_label,
				     (struct disklabel *) data, 0, 
				     slp->sl_dkdev.dk_cpulabel);
		if (error == 0 && cmd == DIOCWDINFO)
		{
			dev_t tdev;

			tdev = MAKEDISKDEV(0, vpdunit(dev), RAW_PART);
			error = writedisklabel(tdev,
					       vpdstrategy, 
					       slp->sl_dkdev.dk_label,
			  		       slp->sl_dkdev.dk_cpulabel);
		}
		slp->sl_flags &= ~VPD_SLICE_LABELLING;
		vpdunlock(sc);
		return error;
	
	case DIOCWLABEL:
		error = suser(p->p_ucred, &p->p_acflag);
		if (error)
			return (error);
		if ((flag & FWRITE) == 0)
			return EBADF;
		if (*(int *) data)
			slp->sl_flags |= VPD_SLICE_WLABEL;
		else
			slp->sl_flags &= ~VPD_SLICE_WLABEL;
		return 0;

	case DIOCGDEFLABEL:
		vpdgetdefaultlabel(sc, (struct disklabel *) data);
		return 0;

	default:
		return ENOTTY;
	}
}


int
vpdsize(dev)
	dev_t dev;
{
	int unit;
	struct vpd_softc *sc;
	struct disklabel *lp;
	struct vpd_slice *slp;
	int size = -1, part = vpdpart(dev);

	unit = vpdunit(dev);
	if (unit >= VPD_NDEVS)
		return size;

	sc = VPD_SOFTC(unit);
	if (sc == NULL || (sc->sc_state != VPD_STATE_OPEN))
		return size;
	
	slp = &sc->sc_sl;
	if ((slp->sl_flags & VPD_SLICE_LOADED) == 0)
	{
		vpdgetdisklabel(sc);
		slp->sl_flags |= VPD_SLICE_LOADED;
	}
	
	lp = slp->sl_dkdev.dk_label;
	if (part == RAW_PART || part == RAW_PART - 1)
		return size;

	if (lp->d_partitions[part].p_fstype == FS_SWAP &&
	    lp->d_partitions[part].p_offset >= 1) 
		size = lp->d_partitions[part].p_size * (lp->d_secsize / dbtob(1));

	return size;
}

int
vpddump(dev, blkno, va, size)
	dev_t dev;
	daddr_t blkno;
	caddr_t va;
	size_t size;
{

	return ENXIO;
}

/*******************************************************************
 * Disklabel 
 *******************************************************************/
static int vpd_read_lower_vnode __P((struct vpd_lower_vnode *, struct proc *));
static struct vpd_lower_vnode *vpd_open_lower_vnode __P((dev_t));
static int vpd_close_lower_vnode __P((struct vpd_lower_vnode *));
static struct vpd_lower_vnode *vpd_allocate_lower_vnode __P((dev_t));

static int
vpd_read_lower_vnode(vup, p)
	struct vpd_lower_vnode *vup;
	struct proc *p;
{
	struct disklabel *lp;
	struct vnode *vp;
	struct buf *bp;
	struct dos_check_args dos;
	int error, i, size;

	vp = vup->vu_vp;
	lp = &vup->vu_d;
	error = VOP_IOCTL(vp, DIOCGDINFO, (caddr_t) lp, FREAD, NOCRED, p);
	if (error)
		return error;

#ifdef	VPD_DISKLABLE_HEADER_INITIALIZE
	lp->d_type = DTYPE_VND;
	lp->d_flags = 0;
	strncpy(lp->d_typename, "vpd disk", 16);
	strncpy(lp->d_packname, "fictitious", 16);
#endif	/* VPD_DISKLABEL_HEADER_INITIALIZE */
	for (i = 0; i < MAXPARTITIONS; i++)
	{
		if (i == RAW_PART || i == RAW_PART - 1)
			continue;
		lp->d_partitions[i].p_offset = 0;
		lp->d_partitions[i].p_size = 0;
		lp->d_partitions[i].p_fstype = FS_UNUSED;
	}

	size = ((int) lp->d_secsize) < NBPG ? NBPG : lp->d_secsize;
	bp = geteblk(size);
	bp->b_blkno = DOSBBSECTOR;
	bp->b_cylin = DOSBBSECTOR / lp->d_secpercyl;
	bp->b_bcount = size;
	bp->b_flags |= B_BUSY | B_READ | B_NOCACHE;
	bgetvp(vp, bp);
	VOP_STRATEGY(bp);
	error = biowait(bp);
	if (error != 0)
		goto out;

	memset(&dos, 0, sizeof(dos));
	dos.lp = lp;
	dos.cdp = &vup->vu_cd;
	dos.bufp = bp->b_data;
	dos.start = NULL;
	dos.dev = vup->vu_tdev;
	error = setup_cpudisklabel(&dos);

out:
	bp->b_flags |= B_INVAL;
	brelse(bp);
	return error;
}

int
vpdgetdisklabel(sc)
	struct vpd_softc *sc;
{
	struct vpd_slice *slp = &sc->sc_sl;
	dev_t dev;
	char *s;

	dev = MAKEDISKDEV(0, sc->sc_dev.dv_unit, RAW_PART);
	s = readdisklabel(dev, vpdstrategy, 
			  slp->sl_dkdev.dk_label,
			  slp->sl_dkdev.dk_cpulabel);
	return 0;
}

/*******************************************************************
 * Target vnode open
 *******************************************************************/
static struct vpd_lower_vnode *
vpd_open_lower_vnode(dev)
	dev_t dev;
{
	struct vpd_lower_vnode *vup;

	for (vup = LIST_FIRST(&vu_tab); vup != NULL; 
	     vup = LIST_NEXT(vup, vu_chain))
	{
		if (vup->vu_tdev == dev)
			break;
	}

	if (vup == NULL)
	{
		vup = vpd_allocate_lower_vnode(dev);
		if (vup == NULL)
			return vup;
	}

	vup->vu_refc ++;
	return vup;
}

static int
vpd_close_lower_vnode(vup)
	struct vpd_lower_vnode *vup;
{
	struct vnode *vp = vup->vu_vp;

	vup->vu_refc --;
	if (vup->vu_refc != 0)
		return 0;

	LIST_REMOVE(vup, vu_chain);

	vn_lock(vp, LK_EXCLUSIVE | LK_RETRY);
	VOP_CLOSE(vp, vup->vu_openf, NOCRED, 0);
	vput(vp);

	free(vup, M_DEVBUF);
	return 0;
}


static struct vpd_lower_vnode *
vpd_allocate_lower_vnode(dev)
	dev_t dev;
{
	struct vpd_lower_vnode *vup;
	struct vnode *tv;
	int error;

	vup = malloc(sizeof(*vup), M_DEVBUF, M_WAITOK);
	bzero(vup, sizeof(*vup));

	vup->vu_tdev = dev;
	if (bdevvp(dev, &tv))
		goto bad;

	vup->vu_openf = FREAD | FWRITE;
	error = VOP_OPEN(tv, vup->vu_openf, NOCRED, 0);
	if (error != 0)
	{
		if (error == EROFS)
		{
			vup->vu_openf = FREAD;
			error = VOP_OPEN(tv, vup->vu_openf, NOCRED, 0);
		}

		if (error != 0)
		{
			vput(tv);
			goto bad;
		}
	}

	VOP_UNLOCK(tv, 0);
	vup->vu_vp = tv;
	error = vpd_read_lower_vnode(vup, 0);
	if (error != 0)
	{
		vn_lock(tv, LK_EXCLUSIVE | LK_RETRY);
		VOP_CLOSE(tv, vup->vu_openf, NOCRED, 0);
		vput(tv);
		goto bad;
	}

	LIST_INSERT_HEAD(&vu_tab, vup, vu_chain);
	return vup;

bad:
	free(vup, M_DEVBUF);
	return NULL;
}

/*******************************************************************
 * Target vnode open/close
 *******************************************************************/
static int
vpdgetdefaultlabel(sc, tlp)
	struct vpd_softc *sc;
	struct disklabel *tlp;
{
	struct vpd_lower_vnode *vup;
	struct cpu_disklabel *osdep;
	int slno;

	vup = sc->sc_vup;
	slno = sc->sc_slno;
 	osdep = &vup->vu_cd;
	*tlp = vup->vu_d;
	tlp->d_partitions[RAW_PART - 1].p_offset = osdep->cd_se[slno].se_start;
	tlp->d_partitions[RAW_PART - 1].p_size = osdep->cd_se[slno].se_size;
	tlp->d_partitions[RAW_PART - 1].p_fstype = FS_UNUSED;
	return 0;
}

static int
vpdvnodeopen(sc, tdev, slno)
	struct vpd_softc *sc;
	dev_t tdev;
	int slno;
{
	struct vpd_slice *slp = &sc->sc_sl;
	struct vpd_lower_vnode *vup;
	struct cpu_disklabel *osdep;
	struct disklabel *lp;
	int oslno;

	if (tdev == 0)
		return ENXIO;

#ifdef	DIAGNOSTIC
	if (sc->sc_vup != NULL)
	{
		printf("%s: already vup allocated\n", sc->sc_dev.dv_xname);
		return EINVAL;
	}
#endif	/* DIAGNOSTIC */

	vup = vpd_open_lower_vnode(tdev);
	if (vup == NULL)
		return ENXIO;

 	osdep = &vup->vu_cd;
	if (slno >= osdep->cd_nslices)
	{
		vpd_close_lower_vnode(vup);
		return EINVAL;
	}

	if (sc->sc_maxnbuf == 0)
	{
		disk_attach(&slp->sl_dkdev);
		sc->sc_maxnbuf = VPD_NMAXIO;
		vpd_init_ccbque(sc->sc_maxnbuf);
	}

	if ((vup->vu_openf & FWRITE) == 0)
		sc->sc_flags |= VPD_RDONLY;
	else
		sc->sc_flags &= ~VPD_RDONLY;

	slp->sl_dkdev.dk_openmask = 0;
	slp->sl_dkdev.dk_copenmask = 0;
	slp->sl_dkdev.dk_bopenmask = 0;
	slp->sl_flags = VPD_SLICE_ACTIVE;
	slp->sl_dsp = disk_slice_hook_insert(
	    MAKEDISKDEV(findblkmajor(sc->sc_dev.dv_cfdata->cf_driver->cd_name),
	    sc->sc_dev.dv_unit, RAW_PART), osdep->cd_se[slno].se_start);
	memset(slp->sl_dkdev.dk_cpulabel, 0, sizeof(struct cpu_disklabel));

	sc->sc_tdev = tdev;
	sc->sc_slno = slno;
	sc->sc_vup = vup;

	lp = slp->sl_dkdev.dk_label;
	vpdgetdefaultlabel(sc, lp);

	oslno = netbsd_start_slice(tdev, osdep);
	if (oslno == slno)
		slp->sl_flags |= VPD_SLICE_BLOCKED;

#ifdef	VPD_PROTECT_OTHERS
	slp->sl_sblk = lp->d_secpercyl * (lp->d_secsize / dbtob(1));
	slp->sl_plblk = lp->d_partitions[RAW_PART - 1].p_offset * 
			(lp->d_secsize / dbtob(1));
	slp->sl_phblk = slp->sl_plblk + 
			(lp->d_partitions[RAW_PART - 1].p_size * 
			(lp->d_secsize / dbtob(1)));
#endif	/* VPD_PROTECT_OTHERS */

	sc->sc_state = VPD_STATE_OPEN;
	return 0;
}

static int
vpdvnodeclose(sc)
	struct vpd_softc *sc;
{
	int s;

	if (sc->sc_vup == NULL)
		return 0;

	s = splbio();
	sc->sc_state = VPD_STATE_CLOSING;
	while (sc->sc_nbuf > 0)
	{
		tsleep((caddr_t) &sc->sc_state, PRIBIO, "vpdclose", 0);
	}
	splx(s);

	vpd_close_lower_vnode(sc->sc_vup);
	sc->sc_state = VPD_STATE_ZERO;
	sc->sc_vup = NULL;
	return 0;
}

/*******************************************************************
 * Auto configuration 
 *******************************************************************/
static char *vpd_lower_device_name[] = { "wd", "sd", NULL };
void vpdautoconf __P((void));

static dev_t
vpdgetbdev(name, unit)
	char *name;
{
	char **np = vpd_lower_device_name;
	dev_t mjr;

	for ( ; *np != NULL && strcmp(*np, name) != 0; np ++)
		;
	if (*np == NULL)
		return 0;

	mjr = findblkmajor(*np);
	if (mjr == (dev_t) -1)
		return 0;

	return (MAKEDISKDEV(mjr, unit, RAW_PART));
}

void
vpdautoconf()
{
	struct device *dv;
	struct vpd_softc *sc;
	struct vpdsrv_softc *dsc;
	struct vpd_lower_vnode *vup;
	int nvpd, i;
	dev_t dev;

	if (vpdsrv_boot == NULL)
		return;
	dsc = vpdsrv_boot;

	for (dv = alldevs.tqh_first, nvpd = 0;
	     dv != NULL && nvpd < dsc->sc_nvpd; dv = dv->dv_list.tqe_next)
	{
		if (dv->dv_class != DV_DISK)
			continue;

		dev = vpdgetbdev(dv->dv_cfdata->cf_driver->cd_name,
				 dv->dv_unit);
		if (dev == 0)
			continue;

		vup = vpd_open_lower_vnode(dev);
		if (vup == NULL)
			continue;

		for (i = 0; i < VPD_NSLICES && nvpd < dsc->sc_nvpd; i ++)
		{
			sc = dsc->sc_vpd[nvpd];
			if (vpdvnodeopen(sc, dev, i) != 0)
				continue;
			vpdvnodeclose(sc);
			strncpy(sc->sc_xname,
				dv->dv_cfdata->cf_driver->cd_name,
				sizeof(sc->sc_xname));
			printf("%s: connecting %s%ds%d\n",
				sc->sc_dev.dv_xname, 
				dv->dv_cfdata->cf_driver->cd_name,
				dv->dv_unit, sc->sc_slno);
			nvpd ++;
		}
		vpd_close_lower_vnode(vup);
	}
}
