/*	$NecBSD: main.c,v 1.1.2.8 2000/07/06 16:41:50 root Exp $	*/
/*	$NetBSD: main.c,v 1.15 1999/02/12 05:14:22 cjs Exp $	*/

#ifndef	ORIGINAL_CODE
/*
 * [NetBSD for NEC PC-98 series]
 *  Copyright (c) 1998, 1999
 *	NetBSD/pc98 porting staff. All rights reserved.
 */
#endif	/* PC-98 */
/*
 * Copyright (c) 1996, 1997
 * 	Matthias Drochner.  All rights reserved.
 * Copyright (c) 1996, 1997
 * 	Perry E. Metzger.  All rights reserved.
 * Copyright (c) 1997
 *	Jason R. Thorpe.  All rights reserved
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgements:
 *	This product includes software developed for the NetBSD Project
 *	by Matthias Drochner.
 *	This product includes software developed for the NetBSD Project
 *	by Perry E. Metzger.
 * 4. The names of the authors may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#ifdef	ORIGINAL_CODE
/* extract from netbsd_pc98:sys/arch/i386/boot/boot.c */
#endif	/* PC-98 */


#include <sys/reboot.h>

#include <lib/libsa/stand.h>
#include <lib/libkern/libkern.h>

#include <libi386.h>
#ifndef	ORIGINAL_CODE
#include "../lib/bootinfo.h"
#include "../lib/bootinfo_bootflags.h"
#include "../lib/disk_direct.h"

extern int devswpos;
#endif	/* PC-98 */

extern void ls __P((char*));
extern int bios2dev __P((int, char**, int*));

int errno;
extern int boot_biosdev;

extern	char bootprog_name[], bootprog_rev[], bootprog_date[],
	bootprog_maker[];

char *names[] = {
    "netbsd", "netbsd.gz",
    "netbsd.old", "netbsd.old.gz",
    "onetbsd", "onetbsd.gz",
#ifdef notyet    
    "netbsd.el", "netbsd.el.gz",
#endif /*notyet*/
};

#define NUMNAMES (sizeof(names)/sizeof(char *))
#define DEFFILENAME names[0]

#define MAXDEVNAME 16

#define TIMEOUT 5

static char *default_devname;
static int default_unit, default_partition;
static char *default_filename;

void	command_help __P((char *));
void	command_ls __P((char *));
void	command_quit __P((char *));
void	command_boot __P((char *));
void	command_dev __P((char *));
#ifndef	ORIGINAL_CODE
void	command_devsw __P((char *));
void	command_bootflags __P((char *));
#endif	/* PC-98 */

struct bootblk_command commands[] = {
	{ "help",	command_help },
	{ "?",		command_help },
	{ "ls",		command_ls },
	{ "quit",	command_quit },
	{ "boot",	command_boot },
	{ "dev",	command_dev },
#ifndef	ORIGINAL_CODE
	{ "devsw",	command_devsw },
	{ "bootflags",	command_bootflags },
#endif	/* PC-98 */
	{ NULL,		NULL },
};

int
parsebootfile(fname, fsname, devname, unit, partition, file)
	const char *fname;
	char **fsname; /* out */
	char **devname; /* out */
	unsigned int *unit, *partition; /* out */
	const char **file; /* out */
{
	const char *col, *help;

	*fsname = "ufs";
	*devname = default_devname;
	*unit = default_unit;
	*partition = default_partition;
	*file = default_filename;

	if (fname == NULL)
		return(0);

	if((col = strchr(fname, ':'))) {	/* device given */
		static char savedevname[MAXDEVNAME+1];
		int devlen;
		unsigned int u = 0, p = 0;
		int i = 0;

		devlen = col - fname;
		if (devlen > MAXDEVNAME)
			return(EINVAL);

#define isvalidname(c) ((c) >= 'a' && (c) <= 'z')
		if (!isvalidname(fname[i]))
			return(EINVAL);
		do {
			savedevname[i] = fname[i];
			i++;
		} while (isvalidname(fname[i]));
		savedevname[i] = '\0';

#define isnum(c) ((c) >= '0' && (c) <= '9')
		if (i < devlen) {
			if (!isnum(fname[i]))
				return(EUNIT);
			do {
				u *= 10;
				u += fname[i++] - '0';
			} while (isnum(fname[i]));
		}

#define isvalidpart(c) ((c) >= 'a' && (c) <= 'z')
		if (i < devlen) {
			if (!isvalidpart(fname[i]))
				return(EPART);
			p = fname[i++] - 'a';
		}

		if (i != devlen)
			return(ENXIO);

		*devname = savedevname;
		*unit = u;
		*partition = p;
		help = col + 1;
	} else
		help = fname;

	if (*help)
		*file = help;

	return(0);
}

char *sprint_bootsel(filename)
const char *filename;
{
	char *fsname, *devname;
	int unit, partition;
	const char *file;
	static char buf[80];
	
	if (parsebootfile(filename, &fsname, &devname, &unit,
			  &partition, &file) == 0) {
		sprintf(buf, "%s%d%c:%s", devname, unit, 'a' + partition, file);
		return(buf);
	}
	return("(invalid)");
}

void
bootit(filename, howto, tell)
	const char *filename;
	int howto, tell;
{

	if (tell) {
		printf("booting %s", sprint_bootsel(filename));
		if (howto)
			printf(" (howto 0x%x)", howto);
		printf("\n");
	}

	if (exec_netbsd(filename, 0, howto) < 0)
		printf("boot: %s: %s\n", sprint_bootsel(filename),
		       strerror(errno));
	else
		printf("boot returned\n");
}

void
print_banner(void)
{

	printf("\n");
	printf(">> %s, Revision %s\n", bootprog_name, bootprog_rev);
	printf(">> (%s, %s)\n", bootprog_maker, bootprog_date);
	printf(">> Memory: %d/%d k\n", getbasemem(), getextmem());
	printf(
#ifdef	ORIGINAL_CODE
#ifdef COMPAT_OLDBOOT
	       "Use hd1a:netbsd to boot sd0 when wd0 is also installed\n"
#endif
#else	/* PC-98 */
		/* XXX: DON'T USE `hd' disk selection */
	       "use sd0a:netbsd to boot sd0\n"
#endif	/* PC-98 */
	       "Press return to boot now, any other key for boot menu\n");
}


/* 
 * note: normally, void main() wouldn't be legal, but this isn't a
 * hosted environment...
 */
int
main()
{
#ifndef	ORIGINAL_CODE
	u_long biosflags;
#endif	/* PC-98 */
	int currname;
	char c;

#ifdef	ORIGINAL_CODE
#ifdef	SUPPORT_SERIAL
	initio(SUPPORT_SERIAL);
#else
	initio(CONSDEV_PC);
#endif
#else	/* PC-98 */

	biosflags = get_biosparam(BOOTFLAGS_BIOS_OFFSET);
#ifdef	SUPPORT_SERIAL
	if ((biosflags & BOOTFLAGS_FLAGS_CONSOLE_SERIAL) != 0)
	{
		initio(CONSDEV_COM0);
	}
	else	
	{
#ifdef	SERIAL_BOOTKEY
		u_long congetshift __P((void));
		
		if ((congetshift() & 0x06) == 0x06)	/* caps & kana on */
			initio(CONSDEV_AUTO);
		else
			initio(CONSDEV_PC);
#else	/* !SERIAL_BOOTKEY */
		initio(CONSDEV_AUTO);
#endif	/* !SERIAL_BOOTKEY */
	}
#else	/* !SUPOORT_SERIAL */
	initio(CONSDEV_PC);
#endif	/* !SUPPORT_SERIAL */

	if ((biosflags & BOOTFLAGS_FLAGS_BOOT_NETWORK) != 0 && ndevs > 1)
	{
		devswpos = 1;
	} 
#endif	/* PC-98 */

	gateA20();

	print_banner();

	/* try to set default device to what BIOS tells us */
	bios2dev(boot_biosdev, &default_devname, &default_unit);
	default_partition = 0;

	/* if the user types "boot" without filename */
	default_filename = DEFFILENAME;

#ifndef	ORIGINAL_CODE
	printf("IPL dev 0x%lx start cyl 0x%lx start ocyl 0x%lx\n",
		IPL_bootdevice, IPL_bootcylinder, IPL_obootcylinder);
	bi_getbootflags();

#ifdef	BOOT_IPPI
	{
		extern void ippiattach __P((void));

		ippiattach();
	}
#endif	/* BOOT_IPPI */

#ifdef	BOOT_DISK_DIRECT
  	{
		int biosdev;

		biosdev = disk_direct_primary(boot_biosdev);
		if (biosdev != boot_biosdev)
		{
			boot_biosdev = biosdev;
			bios2dev(boot_biosdev, &default_devname, &default_unit);
		}
  	}
#endif	/* BOOT_DISK_DIRECT */
#endif	/* PC-98 */

	currname = 0;
	for (;;) {
		printf("booting %s - starting in ",
		       sprint_bootsel(names[currname]));

		c = awaitkey(TIMEOUT, 1);
		if ((c != '\r') && (c != '\n') && (c != '\0')) {
			printf("type \"?\" or \"help\" for help.\n");
			bootmenu(); /* does not return */
		}

		/*
		 * try pairs of names[] entries, foo and foo.gz
		 */
		/* don't print "booting..." again */
		bootit(names[currname], 0, 0);
		/* since it failed, try switching bootfile. */
		currname = ++currname % NUMNAMES;

		/* now try the second of a pair, presumably the .gz
		   version. */
		/* XXX duped code sucks. */
		bootit(names[currname], 0, 1);
		/* since it failed, try switching bootfile. */
		currname = ++currname % NUMNAMES;
	}

	return (0);
}

/* ARGSUSED */
void
command_help(arg)
	char *arg;
{

	printf("commands are:\n"
	    "boot [xdNx:][filename] [-adrs]\n"
	    "     (ex. \"sd0a:netbsd.old -s\"\n"
	    "ls [path]\n"
	    "dev xd[N[x]]:\n"
	    "help|?\n"
#ifndef	ORIGINAL_CODE
	    "devsw    -- change network or local disk device\n"
	    "bootflags\n"
#endif	/* PC-98 */
	    "quit\n");
}

void
command_ls(arg)
	char *arg;
{
	char *save = default_filename;

	default_filename = "/";
	ls(arg);
	default_filename = save;
}

/* ARGSUSED */
void
command_quit(arg)
	char *arg;
{

	printf("Rebooting... goodbye...\n");
	delay(1000000);
	reboot();
	/* Note: we shouldn't get to this point! */
	panic("Could not reboot!");
	exit();
}

void
command_boot(arg)
	char *arg;
{
	char *filename;
	int howto;

	if (parseboot(arg, &filename, &howto))
		bootit(filename, howto, 1);
}

void
command_dev(arg)
	char *arg;
{
	static char savedevname[MAXDEVNAME + 1];
	char *fsname, *devname;
	const char *file; /* dummy */

	if (*arg == '\0') {
		printf("%s%d%c:\n", default_devname, default_unit,
		       'a' + default_partition);
		return;
	}

	if (!strchr(arg, ':') ||
	    parsebootfile(arg, &fsname, &devname, &default_unit,
			  &default_partition, &file)) {
		command_help(NULL);
		return;
	}
	    
	/* put to own static storage */
	strncpy(savedevname, devname, MAXDEVNAME + 1);
	default_devname = savedevname;
}

#ifndef	ORIGINAL_CODE
void
command_devsw(arg)
	char *arg;
{

	devswpos ++;
	if (devswpos >= ndevs)
		devswpos = 0;
	printf("current device: %s\n", DEV_NAME(&devsw[devswpos]));
}

#define	BOOTFLAGS_ARGSLEN	128
static char *get_next_arg __P((char *, char *, int));
static u_long s2num __P((char *, u_long));
static void s2eth __P((char *, struct bootflags_device *));

static char *
get_next_arg(arg, bufp, len)
	char *arg, *bufp;
	int len;
{
	int i;

	bufp[0] = 0;
	for ( ; *arg == ' '; arg ++)
		;
	if (*arg == 0)
		return arg;

	for (i = 0; i < len - 1; i ++, arg ++)
	{
		bufp[i] = *arg;
		if (*arg == ' ' || *arg == 0)
			break;
	}
	bufp[i] = 0;
	for ( ; *arg == ' '; arg ++)
		;
	return arg;
}

static void
s2eth(bp, bdp)
	char *bp;
	struct bootflags_device *bdp;
{
	int i, brk;
	char *obp;
	u_long val;

	for (brk = i = 0; i < 6 && brk == 0; i ++, bp ++)
	{
		obp = bp;
		for ( ; *bp != ':' && *bp != 0; bp ++)
			;
		if (*bp == 0)
			brk = 1;
		else 
			*bp = 0;
		if (obp == bp)
			val = 0;
		else
			val = s2num(obp, 16);
		bdp->bd_loc[i] = val;
	}
}	


static u_long
s2num(bp, card)
	char *bp;
	u_long card;
{
	u_long val = 0;

	for ( ;*bp != 0; bp ++)
	{
		val *= card;
		if (*bp == 'x')
		{
			card = 16;
		}
		else if (card == 16 && *bp >= 'a' && *bp <= 'f')
		{
			val += (u_long) ((*bp) - 'a' + 10);
		}
		else if (*bp >= '0' && *bp <= '9')
		{
			val += (u_long) ((*bp) - '0');
		}
		else
		{
			break;
		}
	}
	return val;
}

static struct bootflags_device_locator iomem_bootflags_locator[] = {
	{ "sel", 0},
	{ "io", 2},
	{ "iomem", 4},
	{ "iosiz", 5},
	{ "irq", 6},
	{ "drq", 7},
	{ NULL, 0}
};
		
void
command_bootflags(arg)
	char *arg;
{
	static char cmdbuf[BOOTFLAGS_ARGSLEN];
	static char argbuf[BOOTFLAGS_ARGSLEN];
	struct bootflags_device bbuf;
	struct bootflags_device *nbdp, *bdp = &bbuf;
	struct bootflags_device_locator *blp, *nblp;
	int i;

	bzero(bdp, sizeof(*bdp));
	for (i = 0; i < 8; i ++)
		bdp->bd_loc[i] = -1;
	bdp->bd_flags = -1;

	nblp = &iomem_bootflags_locator[0];
	while (*arg != 0)
	{
		arg = get_next_arg(arg, cmdbuf, BOOTFLAGS_ARGSLEN);
		arg = get_next_arg(arg, argbuf, BOOTFLAGS_ARGSLEN);
	
		if (strcmp(cmdbuf, "show") == 0)
		{
			show_bootflags_device(NULL, NULL);
			return;
		}

		if (strcmp(cmdbuf, "bus") == 0)
		{
#if	0
			if (strstr(argbuf, "isa") == 0)
				bdp->bd_type |= BOOTFLAGS_DEVICE_TYPE_BUS_ISA;
#endif
			if (strncmp(argbuf, "boot", 4) == 0)
				bdp->bd_type |= BOOTFLAGS_DEVICE_TYPE_TARGET_BOOT;
			continue;
		}


		if (strcmp("name", cmdbuf) == 0)
		{
			strncpy(bdp->bd_name, argbuf, sizeof(bdp->bd_name) - 1);
			if (pmatch(bdp->bd_name, "*.eth.addr", NULL) != 0)
			{
				nblp = NULL;
				arg = get_next_arg(arg, argbuf,
						   BOOTFLAGS_ARGSLEN);
				s2eth(argbuf, bdp);
			}
			continue;
		}

		if (nblp != &iomem_bootflags_locator[0])
		{
			printf("bootflags: %s invalid in the context\n", cmdbuf);
			return;
		}

		if (strcmp("flags", cmdbuf) == 0)
		{
			bdp->bd_flags = s2num(argbuf, 10);
			continue;
		}	

		for (blp = nblp; blp->bl_name != NULL; blp ++)
		{
			if (strcmp(blp->bl_name, cmdbuf) == 0)
			{
				bdp->bd_loc[blp->bl_loc] = s2num(argbuf, 10);
				break;
			}
		}
	}

	if (update_bootflags_device(bdp) != 0)
		goto bad;

	nbdp = lookup_bootflags_device(bdp->bd_type, bdp->bd_name);
	if (nbdp == NULL)
		goto bad;

	show_bootflags_device(nbdp, nblp);
	return;

bad:
	printf("bootflags: not registered\n");
}
#endif	/* PC-98 */
