/*
 * [NetBSD for NEC PC-98 series]
 *  Copyright (c) 2000
 *	NetBSD/pc98 porting staff. All rights reserved.
 * 
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. The name of the author may not be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/types.h>
#include <machine/cpu.h>

#include <lib/libkern/libkern.h>
#include <lib/libsa/stand.h>

#include "libi386.h"
#include "bootinfo.h"
#include "bootinfo_bootflags.h"

struct btinfo_common *bootflags_common_p;
int bootflags_size;

void
bi_getbootflags()
{
	static int only_once;
	int fd;
	struct bootflags_info_header bih;
	struct bootflags_btinfo_header *bbhp;
	struct bootflags_device *bdp;
	u_int8_t *bp;
	u_int32_t bootflags;
	int nent = 0, size, asize;
	u_char *path = "/etc/bootflags.db";

	if (only_once != 0)
		return;
	only_once = 1;

	fd = open(path,  0);
	if (fd < 0)
	{
		printf("bi_getbootflags: %s not found\n", path);
		fd = -1;
	}
	else if (read(fd, &bih, sizeof(bih)) != sizeof(bih))
	{
		printf("bi_getbootflags: header data read failed\n");
		close(fd);
		fd = -1;
	}

	if (fd >= 0)
	{
		if (bih.bi_signature == BOOTFLAGS_SIGNATURE)
		{
			nent = bih.bi_nentry;
			if (nent >= BOOTFLAGS_BTINFO_MAXENTRY)
				nent = BOOTFLAGS_BTINFO_MAXENTRY;
		}
		else
		{
			close(fd);
			fd = -1;
			printf("bi_getbootflags: signature mismatch\n");
		}
	}

	asize = sizeof(*bbhp) + (BOOTFLAGS_BTINFO_MAXENTRY * sizeof(*bdp));
	size = sizeof(*bbhp) + (nent * sizeof(*bdp));
	bp = alloc(asize);
	if (bp == NULL)
		goto out;

	bzero(bp, asize);
	bbhp = (void *) bp;
	bdp = (void *) (bp + sizeof(*bbhp));
	if (fd >= 0 && nent > 0)
	{
		if (read(fd, bp + sizeof(*bbhp), nent * sizeof(*bdp)) 
		    != nent * sizeof(*bdp))
		{
			close(fd);
			fd = -1;
			nent = 0;
			printf("bi_getbootflags: read failed\n");
		}
	}

	bbhp->bb_bi.bi_signature = BOOTFLAGS_SIGNATURE;
	bbhp->bb_bi.bi_nentry = nent;
	bootflags = get_biosparam(BOOTFLAGS_BIOS_OFFSET);
	bbhp->bb_bootflags = bootflags;
	/* BI_ADD(&bbhp->bb_common, BTINFO_BOOTFLAGS, size); */
	printf("bi_getbootflags: bootflags succeeded(%d)\n", nent);
	bootflags_common_p = &bbhp->bb_common;
	bootflags_size = size;

out:
	if (fd >= 0)
		close(fd);
}

#ifdef	BOOTINFO_BOOTFLAGS_EXT
static void __show_bootflags_device __P((struct bootflags_device *, struct bootflags_device_locator *));

struct bootflags_device *
lookup_bootflags_device(type, s)
	u_long type;
	u_char *s;
{
	u_int8_t *bp = (u_int8_t *) bootflags_common_p;
	struct bootflags_btinfo_header *bbhp;
	struct bootflags_device *bdp;
	int nent, i;

	if (bp == NULL)	/* not found */
		return NULL;

	bbhp = (void *) bp;
	bdp = (void *) (bp + sizeof(*bbhp));
 	nent = bbhp->bb_bi.bi_nentry;

	for (i = 0; i < nent; i ++, bdp ++)
	{
		if (bdp->bd_type == type &&
		    strncmp(bdp->bd_name, s, sizeof(bdp->bd_name)) == 0)
			return bdp;
	}
	return NULL;
}

static u_char *delete_mark = "unused";

int
delete_bootflags_device(type, s)
	u_long type;
	u_char *s;
{
	struct bootflags_device *bdp;

	bdp = lookup_bootflags_device(type, s);
	if (bdp == NULL)
		return ENOENT;

	strncpy(bdp->bd_name, delete_mark, sizeof(bdp->bd_name));
	bdp->bd_type = -1;
	return 0;
}

int
update_bootflags_device(nbdp)
	struct bootflags_device *nbdp;
{
	u_int8_t *bp = (u_int8_t *) bootflags_common_p;
	struct bootflags_btinfo_header *bbhp;
	struct bootflags_device *bdp;
	int size;

	if (bp == NULL)
		return ENOENT;
	if (nbdp->bd_name[0] == 0)
		return EINVAL;

	bbhp = (void *) bp;
	bdp = lookup_bootflags_device(nbdp->bd_type, nbdp->bd_name);
	if (bdp != NULL)
		goto found;

	bdp = lookup_bootflags_device(-1, delete_mark);
	if (bdp != NULL)
		goto found;

	if (bbhp->bb_bi.bi_nentry >= BOOTFLAGS_BTINFO_MAXENTRY)
		return ENOSPC;

	size = bootflags_size;
	bdp = (void *) (bp + size);
	bootflags_size = size + sizeof(*bdp);
	bbhp->bb_bi.bi_nentry ++;

found:	
	bcopy(nbdp, bdp, sizeof(*bdp));
	return 0;
}


static struct bootflags_device_locator default_locator[] = {
	{ "loc0", 0},
	{ "loc1", 1},
	{ "loc2", 2},
	{ "loc3", 3},
	{ "loc4", 4},
	{ "loc5", 5},
	{ "loc6", 6},
	{ "loc7", 7},
	{ NULL,}
};

static void
__show_bootflags_device(bdp, blp)
	struct bootflags_device *bdp;
	struct bootflags_device_locator *blp;
{
	u_long val;

	if (blp == NULL)
		blp = &default_locator[0];

	printf("[*] type(0x%lx) name(%s) flags(0x%lx)\n",
		 bdp->bd_type, bdp->bd_name, bdp->bd_flags);

	for (; blp->bl_name != NULL; blp ++)
	{
		val = bdp->bd_loc[blp->bl_loc];
		if (val == (u_long) -1)
			continue;
		printf("%s(0x%lx) ", blp->bl_name, val);
	}
	printf("\n");
}

void
show_bootflags_device(nbdp, blp)
	struct bootflags_device *nbdp;
	struct bootflags_device_locator *blp;
{
	u_int8_t *bp = (u_int8_t *) bootflags_common_p;
	struct bootflags_btinfo_header *bbhp;
	struct bootflags_device *bdp;
	int nent, i;

	if (bp == NULL)
		return;

	bbhp = (void *) bp;
	bdp = (void *) (bp + sizeof(*bbhp));
	nent = bbhp->bb_bi.bi_nentry;

	for (i = 0; i < nent; i ++, bdp ++)
	{
		if (nbdp == NULL || bdp == nbdp)
		{
			__show_bootflags_device(bdp, blp);
		}
        } 
}
#endif	/* BOOTINFO_BOOTFLAGS_EXT */
