/*
 * [NetBSD for NEC PC-98 series]
 *  Copyright (c) 2000
 *	NetBSD/pc98 porting staff. All rights reserved.
 * 
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. The name of the author may not be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/device.h>
#include "busio_user.h"

u_int8_t 
bus_space_read_1(bst, bsh, offs)
	bus_space_tag_t bst;
	bus_space_handle_t bsh;
	bus_size_t offs;
{
	register bus_addr_t port;

	if (bst == NULL)
		port = bsh + offs;
	else
		port = bsh + bst[offs];

	return inb(port);
}

void
bus_space_write_1(bst, bsh, offs, val)
	bus_space_tag_t bst;
	bus_space_handle_t bsh;
	bus_size_t offs;
	u_int8_t val;
{
	register bus_addr_t port;

	if (bst == NULL)
		port = bsh + offs;
	else
		port = bsh + bst[offs];


	outb(port, val);
}

u_int16_t 
bus_space_read_2(bst, bsh, offs)
	bus_space_tag_t bst;
	bus_space_handle_t bsh;
	bus_size_t offs;
{
	register bus_addr_t port;

	if (bst == NULL)
		port = bsh + offs;
	else
		port = bsh + bst[offs];

	return inw(port);
}

void
bus_space_write_2(bst, bsh, offs, val)
	bus_space_tag_t bst;
	bus_space_handle_t bsh;
	bus_size_t offs;
	u_int16_t val;
{
	register bus_addr_t port;

	if (bst == NULL)
		port = bsh + offs;
	else
		port = bsh + bst[offs];

	outw(port, val);
}

void
bus_space_read_multi_2(bst, bsh, offs, bufp, count)
	bus_space_tag_t bst;
	bus_space_handle_t bsh;
	bus_size_t offs;
	u_int16_t *bufp;
	bus_size_t count;
{
	register bus_addr_t port;

	if (bst == NULL)
		port = bsh + offs;
	else
		port = bsh + bst[offs];

	insw(port, bufp, count);
}

void
bus_space_write_multi_2(bst, bsh, offs, bufp, count)
	bus_space_tag_t bst;
	bus_space_handle_t bsh;
	bus_size_t offs;
	u_int16_t *bufp;
	bus_size_t count;
{
	register bus_addr_t port;

	if (bst == NULL)
		port = bsh + offs;
	else
		port = bsh + bst[offs];

	outsw(port, bufp, count);
}
