/*	$NetBSD$	*/
#ifdef	BOOT_PCCS
/*
 * [NetBSD for NEC PC-98 series]
 *  Copyright (c) 2000
 *	NetBSD/pc98 porting staff. All rights reserved.
 * 
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. The name of the author may not be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * Copyright (c) 1995, 1996, 1997, 1998
 *	Naofumi HONDA.  All rights reserved.
 */

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/errno.h>
#include <machine/cpu.h>

#include <i386/Cbus/pccs16/i82365reg.h>
#include "pccs.h"

#define	IOFGSHIFT 4
#define	GIPC_IS_ID_VALID(id) ((id) == 0x82 || (id) == 0x83 || (id) == 0x84)
#define	GIPC_IS_ID_BRIDGE(id) ((id) == 0x84)

struct gipc_softc {
	bus_space_tag_t sc_bst;			/* controller space tag */
	bus_space_handle_t sc_bsh;		/* controller space handle */
};

static __inline void gipc_reg_write_1 __P((pccshw_tag_t, slot_handle_t, u_int, u_int8_t));
static __inline u_int8_t gipc_reg_read_1 __P((pccshw_tag_t, slot_handle_t, u_int));
static __inline void gipc_set_bits 
	__P((pccshw_tag_t, slot_handle_t, u_int, u_int8_t));
static __inline void gipc_clr_bits 
	__P((pccshw_tag_t, slot_handle_t, u_int, u_int8_t));
static int gipcprobesubr __P((bus_space_tag_t, bus_space_handle_t));
static void gipc_open_intr __P((pccshw_tag_t, slot_handle_t));
static int gipc_mapmem __P((pccshw_tag_t, slot_handle_t, int, struct slot_device_iomem *));
static int gipc_mapio __P((pccshw_tag_t, slot_handle_t, int, struct slot_device_iomem *));
static void gipc_reclaim __P((pccshw_tag_t, slot_handle_t));
static int gipc_stat __P((pccshw_tag_t, slot_handle_t, int));
static int gipc_power __P((pccshw_tag_t, slot_handle_t, struct power *, int));
static void gipc_reset __P((pccshw_tag_t, slot_handle_t, int));

/********************************************
 * Bus operations macro
 *******************************************/
/********************************************
 * Probe Attach
 *******************************************/
static int
gipcprobesubr(bst, bsh)
	bus_space_tag_t bst;
	bus_space_handle_t bsh;
{
	u_int8_t val;

	bus_space_write_1(bst, bsh, i82365_index, REG_ID);
	val = bus_space_read_1(bst, bsh, i82365_data); 
	if (GIPC_IS_ID_VALID(val) == 0)
		return 0;

	return ((u_int) val);
}

static struct pccshw_tag pt;
	
int
pccshw_init(iobase)
	u_long iobase;
{
	static struct gipc_softc gs;
	pccshw_tag_t pp = &pt;

	gs.sc_bst = 0;
	gs.sc_bsh = iobase;

	pp->pp_chip = 0;
	pp->pp_dv = &gs;
	pp->pp_maxslots = 4;
	pp->pp_nio = SLOT_DEVICE_NIO;
	pp->pp_nmem = SLOT_DEVICE_NMEM;
	pp->pp_nim = pp->pp_nio + pp->pp_nim;
	pp->pp_ps = 12;
	pp->pp_psz = PCCS_ATTRWIN_SIZE;

	if (gipcprobesubr(gs.sc_bst, gs.sc_bsh) == 0)
		return EINVAL;
	return 0;
}

int
pccshw_slot_init(sid)
	slot_handle_t sid;
{
	pccshw_tag_t pp = &pt;
	struct power pow;

	if (pccshw_init(PCCS_CHIP_IOBASE) != 0)
		return EINVAL;
	if (sid >= pp->pp_maxslots)
		return EINVAL;
	if (gipc_stat(pp, sid, PCCSHW_HASSLOT) == 0)
		return EINVAL;
	if (gipc_stat(pp, sid, PCCSHW_CARDIN) == 0)
		return EINVAL;

	pow.pow_vcc0 = PCCS_POWER_DEFAULT;
	pow.pow_vpp0 = PCCS_POWER_DEFAULT;
	gipc_power(pp, sid, &pow, PCCSHW_POWER_ON);
	gipc_reset(pp, sid, IF_IO);
	gipc_reclaim(pp, sid);
	return 0;
}

int
pccshw_slot_shutdown(sid)
	slot_handle_t sid;
{
	pccshw_tag_t pp = &pt;
	struct power pow;

	gipc_reclaim(pp, sid);
	pow.pow_vcc0 = PCCS_POWER_DEFAULT;
	pow.pow_vpp0 = PCCS_POWER_DEFAULT;
	gipc_power(pp, sid, &pow, PCCSHW_POWER_OFF);
	return 0;
}

int
pccshw_iomem_map(sid, id, wp)
	slot_handle_t sid;
	int id;
	struct slot_device_iomem *wp;
{
	pccshw_tag_t pp = &pt;

	if (id < SLOT_DEVICE_NIO)
		return gipc_mapio(pp, sid, id, wp);
	else if (id < SLOT_DEVICE_NIO + SLOT_DEVICE_NMEM)
		return gipc_mapmem(pp, sid, id, wp);
	else
		return EINVAL;
}

/********************************************
 * Base Func
 *******************************************/
static __inline u_int8_t
gipc_reg_read_1(pp, sid, reg)
	pccshw_tag_t pp;
	slot_handle_t sid;
	u_int reg;
{
	register struct gipc_softc *sc = pp->pp_dv;
	register bus_space_tag_t bst = sc->sc_bst;
	register bus_space_handle_t bsh = sc->sc_bsh;
	bus_size_t offset = sid * REG_SLOT_SKIP + reg;

	bus_space_write_1(bst, bsh, i82365_index, offset);
	return bus_space_read_1(bst, bsh, i82365_data);
}

static __inline void
gipc_reg_write_1(pp, sid, reg, data)
	pccshw_tag_t pp;
	slot_handle_t sid;
	u_int reg;
	u_int8_t data;
{
	register struct gipc_softc *sc = pp->pp_dv;
	register bus_space_tag_t bst = sc->sc_bst;
	register bus_space_handle_t bsh = sc->sc_bsh;
	bus_size_t offset = sid * REG_SLOT_SKIP + reg;

	bus_space_write_1(bst, bsh, i82365_index, offset);
	bus_space_write_1(bst, bsh, i82365_data, data);
}

static __inline void
gipc_set_bits(pp, sid, reg, mask)
	pccshw_tag_t pp;
	slot_handle_t sid;
	u_int reg;
	u_int8_t mask;
{

	gipc_reg_write_1(pp, sid, reg, gipc_reg_read_1(pp, sid, reg) | mask);
}

static __inline void
gipc_clr_bits(pp, sid, reg, mask)
	pccshw_tag_t pp;
	slot_handle_t sid;
	u_int reg;
	u_int8_t mask;
{

	gipc_reg_write_1(pp, sid, reg, gipc_reg_read_1(pp, sid, reg) & (~mask));
}

/********************************************
 * Power and Reset
 *******************************************/
static int
gipc_power(pp, sid, powp, on)
	pccshw_tag_t pp;
	slot_handle_t sid;
	struct power *powp;
	int on;
{
	u_int8_t bits;

	switch (on)
	{
	case PCCSHW_POWER_ON:
	case PCCSHW_POWER_UPDATE:
		bits = POWR_AUTOPOW | POWR_RESDRV;
		if (powp->pow_vcc0 || powp->pow_vcc0 == PCCS_POWER_DEFAULT)
			bits |= POWR_VCC5V;
		if (powp->pow_vpp0 == 50 || powp->pow_vpp0 == PCCS_POWER_DEFAULT)
			bits |= POWR_VPP5V;
		else if (powp->pow_vpp0 == 120)
			bits |= POWR_VPP12V;

		gipc_reg_write_1(pp, sid, REG_POW, bits);
		pccs_delay(100);
		gipc_set_bits(pp, sid, REG_POW, POWR_OUTEN);
		break;

	case PCCSHW_POWER_OFF:
		gipc_reg_write_1(pp, sid, REG_IGC, IGCR_DISRST);
		gipc_reg_write_1(pp, sid, REG_POW, 0);
		break;

	}

	return 0;
}

static void
gipc_reset(pp, sid, type)
	pccshw_tag_t pp;
	slot_handle_t sid;
	int type;
{
	int tout;

	gipc_reg_write_1(pp, sid, REG_IGC, (type == IF_IO) ? IGCR_IOCARD : 0);
	pccs_delay(100);
	gipc_set_bits(pp, sid, REG_IGC, IGCR_DISRST);
	pccs_delay(100);
	for (tout = 0; (gipc_reg_read_1(pp, sid, REG_STAT) & STAT_RDY)  == 0 &&
			tout < 30; tout ++)
		pccs_delay(100);

	gipc_set_bits(pp, sid, REG_POW, POWR_RESDRV);
	gipc_open_intr(pp, sid);
}

/********************************************
 * Io and Mem Mapping
 *******************************************/
#define	GETFG(topval, shift) ((busflags / (topval)) << (shift))

static void
gipc_reclaim(pp, sid)
	pccshw_tag_t pp;
	slot_handle_t sid;
{

	/* unmap io wins */
	gipc_clr_bits(pp, sid, REG_WE, WE_IOMASK | WE_MEMMASK);
	gipc_set_bits(pp, sid, REG_WE, WE_MEMCS16);

	/* unmap irq & memory card type */
	gipc_clr_bits(pp, sid, REG_IGC, IGCR_IRQM | IGCR_INTR);
}

static int
gipc_mapmem(pp, sid, id, wp)
	pccshw_tag_t pp;
	slot_handle_t sid;
	int id;
	struct slot_device_iomem *wp;
{
	u_int rb, enable, busflags;
	bus_addr_t phys, endphys, offset;

	if (id < SLOT_DEVICE_NIO)
		return EINVAL;

	id -= SLOT_DEVICE_NIO;
	enable = (WE_MEMBASE << id);

	phys = PCCSHW_TRUNC_PAGE(pp, wp->im_hwbase);
	offset = PCCSHW_TRUNC_PAGE(pp, wp->im_base);
	offset = PCCSHW_PHYS2PAGE(pp, offset - phys) & 0x3fff;
	phys = PCCSHW_PHYS2PAGE(pp, phys);
	endphys = PCCSHW_ROUND_PAGE(pp, wp->im_hwbase + wp->im_size);
	endphys = PCCSHW_PHYS2PAGE(pp, endphys) - 1;

	/* setup hw */
	rb = REG_MWBASE + id * REG_MEM_WIN_SKIP;

	busflags = 0;
	if ((wp->im_flags & (SDIM_BUS_WEIGHT | SDIM_BUS_WIDTH16)) == 0)
		busflags |= 0x40;
	if (wp->im_flags & SDIM_BUS_WIDTH16)
		busflags |= 0x80;
	gipc_reg_write_1(pp, sid, rb, phys);
	gipc_reg_write_1(pp, sid, rb + 1, (phys >> 8) | busflags);

	busflags = 0;
	if (wp->im_flags & SDIM_BUS_WEIGHT)
		busflags |= 0xc0;
	gipc_reg_write_1(pp, sid, rb + 2, endphys);
	gipc_reg_write_1(pp, sid, rb + 3, (endphys >> 8) | busflags);

	busflags = 0;
	if (wp->im_flags & WFMEM_ATTR)
		busflags |= 0x40;
	if (wp->im_flags & SDIM_BUS_WP)
		busflags |= 0x80;
	gipc_reg_write_1(pp, sid, rb + 4, offset);
	gipc_reg_write_1(pp, sid, rb + 5, (offset >> 8) | busflags);

	gipc_set_bits(pp, sid, REG_WE, WE_MEMCS16 | enable);
	pccs_delay(50);
	return 0;
}

static int
gipc_mapio(pp, sid, winid, wp)
	pccshw_tag_t pp;
	slot_handle_t sid;
	int winid;
	struct slot_device_iomem *wp;
{
	bus_addr_t phys, endphys;
	u_int rb, busflags, enable, mask;

	if (winid >= SLOT_DEVICE_NIO)
		return EINVAL;

	busflags = 0;
	if (wp->im_flags & SDIM_BUS_WEIGHT)
		busflags |= 0x08;
	if ((wp->im_flags & (SDIM_BUS_WEIGHT | SDIM_BUS_WIDTH16)) == 0)
		busflags |= 0x04;
	if (wp->im_flags & SDIM_BUS_AUTO)
		busflags |= 0x02;
	if (wp->im_flags & SDIM_BUS_WIDTH16)
		busflags |= 0x01;

	phys = wp->im_hwbase;
	endphys = phys + wp->im_size - 1;
	rb = REG_IOWBASE + winid * REG_IO_WIN_SKIP;
	gipc_reg_write_1(pp, sid, rb, phys & 0xff);
	gipc_reg_write_1(pp, sid, rb + 1, (phys >> 8) & 0xff);
	gipc_reg_write_1(pp, sid, rb + 2, endphys & 0xff);
	gipc_reg_write_1(pp, sid, rb + 3, (endphys >> 8) & 0xff);

	/* open windows */
	enable = (WE_IOBASE << winid);
	gipc_set_bits(pp, sid, REG_WE, enable);

	mask = 0x0f << (4 * winid);
	gipc_clr_bits(pp, sid, REG_IOC, mask);

	busflags = busflags << (4 * winid);
	gipc_set_bits(pp, sid, REG_IOC, busflags);

	pccs_delay(50);
	return 0;
}

/********************************************
 * Intr Mapping
 *******************************************/
static void
gipc_open_intr(pp, sid)
	pccshw_tag_t pp;
	slot_handle_t sid;
{

	gipc_clr_bits(pp, sid, REG_IGC, IGCR_IRQM);

	gipc_reg_write_1(pp, sid, REG_CCS, 0);
	(void) gipc_reg_read_1(pp, sid, REG_CS);
	(void) gipc_reg_read_1(pp, sid, REG_STAT);
}

static int
gipc_stat(pp, sid, scmd)
	pccshw_tag_t pp;
	slot_handle_t sid;
	int scmd;
{
	u_int8_t id;

	switch (scmd)	
	{
	case PCCSHW_HASSLOT:
		id = gipc_reg_read_1(pp, sid, REG_ID);
		return GIPC_IS_ID_VALID(id);

	case PCCSHW_CARDIN:
		return ((gipc_reg_read_1(pp, sid, REG_STAT) & STAT_CD) == STAT_CD);

	case PCCSHW_STCHG:
		return (gipc_reg_read_1(pp, sid, REG_CS) & CS_CCHG);

	default:
		return 0;
	}
}
#endif	/* BOOT_PCCS */
