/*	$NetBSD$	*/
/*
 * [NetBSD for NEC PC-98 series]
 *  Copyright (c) 2000
 *	NetBSD/pc98 porting staff. All rights reserved.
 * 
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. The name of the author may not be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */
/* #define	NE_ISA_FAKE_PCCS_DEBUG	1 */

#include <sys/types.h>
#include <sys/param.h>
#include <sys/systm.h>
#include <sys/errno.h>
#include <sys/malloc.h>
#include <sys/device.h>
#include <machine/pio.h>

#include <lib/libsa/stand.h>
#include <libi386.h>

#include <dev/ic/dp8390reg.h>
#include <dev/ic/ne2000reg.h>
#include "../busio_user.h"
#include "dp8390piosubr.h"
#include "ne.h"
#ifdef	NE_ISA_FAKE_PCCS_DEBUG
#include "../pccs.h"
#endif	/* NE_ISA_FAKE_PCCS_DEBUG */

#ifdef _STANDALONE
#include <lib/libkern/libkern.h>
#include <bootinfo.h>
#include "../bootinfo_bootflags.h"
#endif

#include "etherdrv.h"

static struct dp8390pio_softc dp8390_softc_tmp;

#ifdef _STANDALONE
static struct btinfo_netif bi_netif;
#endif

struct nvt_iat_mkinfo {
#define	NE_ISA_CORE_IAT_TBLSZ	0x10
#define	NE_ISA_ASIC_IAT_TBLSZ	0x10
#define	NE_ISA_IAT_FLEN		0x08
#define	NE_ISA_IAT_SLEN		0x08
	bus_size_t	nh_skip;	/* core port offset skip */
	bus_size_t	nh_fgap;	/* core port offset gap */

	bus_addr_t	nh_asic;	/* asic port offset */
	bus_addr_t	nh_data;	/* data port offset */
	bus_addr_t	nh_reset;	/* reset port offset */
};

static bus_size_t ne_isa_core_iat[NE_ISA_CORE_IAT_TBLSZ];
static bus_size_t ne_isa_asic_iat[NE_ISA_ASIC_IAT_TBLSZ];

static struct nvt_iat_mkinfo nehwsel_array[] = {
/* 0x00 */	/* &nehw_pure,*/ { 0x01, 0x08, 0x10, 0x10, 0x1f },
/* 0x01 */	/* &nehw_pure,*/ { 0x01, 0x08, 0x10, 0x10, 0x1f },
/* 0x02 */	/* &nehw_SMARTCOM,*/ { 0x1000, 0x8000, 0x100, 0x100, 0x7f00 },
/* 0x03 */	/* &nehw_EGY98,*/ { 0x02, 0x100, 0x200, 0x200, 0x300 },
/* 0x04 */	/* &nehw_LGY98,*/ { 0x01, 0x08, 0x200, 0x200, 0x300 },
/* 0x05 */	/* &nehw_IF27,*/ { 0x01, 0x08, 0x100, 0x100, 0x10f },
/* 0x06 */	/* &nehw_sic98, */{0, },
/* 0x07 */	/* &nehw_apex, */ {0, },
/* 0x08 */	/* */	          {0, },
/* 0x09 */	/* &nehw_IOLA98,*/ { 0x1000, 0x8000, 0x100, 0x100, 0xf100 },
/* 0x0a */	/* &nehw_cnet98, */ {0, },
/* 0x0b */	/* &nehw_cnet98el, */ {0, },
/* 0x0c */	/* &nehw_NW98X3, */ { 0x0100, 0x0800, 0x1000, 0x1000, 0x1f00 },
};

#ifdef	NE_ISA_FAKE_PCCS_DEBUG
static struct card_info citmp;
static int ne_pccs_fake_open __P((void));

static int
ne_pccs_fake_open()
{
	struct card_info *ci = &citmp;
	slot_handle_t sid = 0;

	bzero(ci, sizeof(*ci));
	if (pccs_open(sid, ci) != 0)
		return 0;

	if (pccs_card_open(sid, ci) != 0)
		return 0;
	return 1;
}
#endif	/* NE_ISA_FAKE_PCCS_DEBUG */

int
ne_isa_init(myadr, argp)
	unsigned char *myadr;
	void **argp;
{
	struct dp8390pio_softc *sc = &dp8390_softc_tmp;
	struct bootflags_device *bdp;
	struct nvt_iat_mkinfo *nvtp;
	u_long iobase, flags;
	bus_addr_t *iat;
	int i;

	bzero(sc, sizeof(*sc));
	bdp = lookup_bootflags_device(BOOTFLAGS_DEVICE_TYPE_TARGET_BOOT |
				      BOOTFLAGS_DEVICE_TYPE_BUS_ISA,
				      "ne");
	if (bdp == NULL)
		return 0;

#ifdef	NE_ISA_FAKE_PCCS_DEBUG
	if (ne_pccs_fake_open() ==0)
		return 0;
#endif	/* NE_ISA_FAKE_PCCS_DEBUG */

	iobase = bdp->bd_loc[2];
	flags = bdp->bd_flags;
	flags = flags >> 20;
	printf("ne_isa: iobase 0x%x flags 0x%x\n",
	       (u_int) iobase, (u_int) flags);
	if (flags >= sizeof(nehwsel_array) / sizeof(*nvtp))
		return 0;

	nvtp = &nehwsel_array[flags];
	if (nvtp->nh_skip == 0)
		return 0;

	iat = ne_isa_core_iat;
	for (i = 0; i < NE_ISA_IAT_FLEN; i++)
		*iat ++ = nvtp->nh_skip * i;
	for (i = 0; i < NE_ISA_IAT_SLEN; i++)
		*iat ++ = nvtp->nh_fgap + nvtp->nh_skip * i;

	iat = ne_isa_asic_iat;
	for (i = 0; i < NE_ISA_ASIC_IAT_TBLSZ; i ++)
		iat[i] = nvtp->nh_asic;
	iat[NE2000_ASIC_DATA] = nvtp->nh_data;
	iat[NE2000_ASIC_RESET] = nvtp->nh_reset;

	sc->sc_nict = ne_isa_core_iat;
	sc->sc_asict = ne_isa_asic_iat;
	sc->sc_nich = iobase;
	sc->sc_asich = iobase;
	sc->cr_proto = ED_CR_RD2;
	sc->sc_useword = 1;
	sc->dcr_reg = ED_DCR_FT1 | ED_DCR_LS | ED_DCR_WTS;
	sc->sc_romstart = 0;

	sc->eth_myaddr[0] = 0xff;
	if (dp8390_config(sc, 16384, 16384) != 0)
		return 0;
	bcopy(sc->eth_myaddr, myadr, 6);

#ifdef _STANDALONE
	bi_netif.addr.iobase = iobase;
	strncpy(bi_netif.ifname, "ne", sizeof(bi_netif.ifname));
	bi_netif.bus = BI_BUS_ISA;
	BI_ADD(&bi_netif, BTINFO_NETIF, sizeof(bi_netif));
#endif
	*argp = sc;
	return 1;
}

void
ne_isa_stop(arg)
	void *arg;
{
	struct dp8390pio_softc *sc = arg;

	dp8390_stop(sc);
}
