/*	$NetBSD$	*/
#ifdef	BOOT_PCCS
/*
 * [NetBSD for NEC PC-98 series]
 *  Copyright (c) 2000
 *	NetBSD/pc98 porting staff. All rights reserved.
 * 
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. The name of the author may not be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/errno.h>
#include <sys/malloc.h>
#include <sys/fcntl.h>
#include <sys/device.h>
#include <sys/proc.h>
#include <machine/cpu.h>

#include "pccs.h"

static u_int pccs_show_info;
static int read_cis_mem __P((u_int8_t *, cis_addr_t, cis_size_t, u_int8_t *));
void bits_expand __P((u_int, u_char *));
static __inline void pccs16_cr_write_1 __P((u_long, u_int, u_int8_t));
static int pccs16_map_mem __P((slot_handle_t, bus_addr_t, int));

struct slot_device_iomem attrwin = {
	0,
	PCCS_ATTRWIN_BASEADDR,
	PCCS_ATTRWIN_SIZE,
	SDIM_BUS_ACTIVE | SDIM_BUS_WEIGHT | WFMEM_ATTR,
	SLOT_DEVICE_SPMEM
};

static u_int8_t tmpbuf[PCCS_ATTRWIN_SIZE + 3];
static cis_addr_t Cisoffset;
u_int8_t *cisbufp;

/* #define	PCCS_DEBUG(mesg)	{printf(mesg);} */
#define	PCCS_DEBUG(mesg)

/************************************************************
 *  DEBUG SECTION
 ************************************************************/
/* #define	USER_SPACE_DEBUG 1 */
#ifdef	USER_SPACE_DEBUG
#include <sys/mman.h>
#include <errno.h>
#undef	PCCS_PHYSTOV
#define	PCCS_PHYSTOV(x)	(vmaddr)

u_char *vmaddr;
u_char buffer[0x10000];

int
main(argc, argv)
	int argc;
	char **argv;
{
	int fd;
	void *addr;
	u_int8_t myadr[12];

	fd = open("/dev/mem", O_RDWR, 0);
	if (fd < 0)
	{
		perror(argv[0]);
		exit(1);
	}

	addr = mmap(0 , PCCS_ATTRWIN_SIZE,
		   PROT_READ | PROT_WRITE, MAP_FILE,
		   fd, (off_t) PCCS_ATTRWIN_BASEADDR);
	if (addr == MAP_FAILED)
	{
		perror(argv[0]);
		exit(1);
	}

	vmaddr = addr;
	i386_iopl(3);
	EtherInit(myadr);
	while(1)
	{
		int len, i;

		len = EtherReceive(buffer, sizeof(buffer));
		if (len > 0)
		{
			for (i = 0; i < len; i ++)
				printf(" 0x%x", (u_int) buffer[i]);
			printf("\n");
		}
	}
	return 0;
}

void
delay(x)
	int x;
{
	sleep(1);
}

void
bits_expand(ul, p)
	u_int ul;
	u_char *p;
{
	int tmp;
	u_char n;

	printf("0x%x", ul);
	p++;
	for (tmp = 0; n = *p++; )
	{
		if (ul & (1 << (n - 1)))
		{
			putchar(tmp ? ',' : '<');
			for (; (n = *p) > ' '; ++p)
				putchar(n);
			tmp = 1;
		}
		else
			for (; *p > ' '; ++p)
				continue;
	}
	if (tmp)
		putchar('>');
	printf("\n");
}
#else	/* !USER_SPACE_DEBUG */
void
bits_expand(ul, p)
	u_int ul;
	u_char *p;
{

	printf("\n");
}
#endif /* !USER_SPACE_DEBUG */
/************************************************************
 * utils
 ************************************************************/
int
pccs_delay(msec)
	int msec;
{

	delay(1024 * msec);
	return 0;
}

static int
read_cis_mem(abuf, offs, len, buf)
	u_int8_t *abuf;
	cis_addr_t offs;
	cis_size_t len;
	u_int8_t *buf;
{
	cis_addr_t offset, page;

	offset = offs & (CARD_ATTR_SZ - 1);
	page = offs - offset;

	if (page != Cisoffset)
		return EINVAL;
	
	if (offset + len >= CARD_ATTR_SZ)
		return EINVAL;

	if (buf != NULL)
		bcopy(abuf + offset, buf, len);
	return 0;
}

int
pccs_open(sid, ci)
	slot_handle_t sid;
	struct card_info *ci;
{
	u_int8_t *tp = (u_int8_t *) PCCS_PHYSTOV(attrwin.im_hwbase);
	int rv, i;

	PCCS_DEBUG("pccs_open: start\r\n")
	rv = pccshw_slot_init(sid);
	if (rv != 0)
		return rv;
	PCCS_DEBUG("pccs_open: slot_init ok\r\n")

	rv = pccshw_iomem_map(sid, SLOT_DEVICE_NIO, &attrwin);
	if (rv != 0)
		goto bad;
	PCCS_DEBUG("pccs_open: attr map ok\r\n")

	for (i = 0; i < PCCS_ATTRWIN_SIZE; i += 2)
		tmpbuf[i / 2] = tp[i];
	Cisoffset = CIS_ATTR_ADDR(0);

	PCCS_DEBUG("pccs_open: parse start\r\n")
	pccs_init_ci(ci);
	PCCS_SET_MFCID(ci, 0);	/* mfcid == 0 */
	pccscis_parse_cis(tmpbuf, ci, PCCS_NOT_FOLLOW_LINK);
	PCCS_DEBUG("pccs_open: parse ok\r\n")

	show_card_info(ci);
	cisbufp = tmpbuf;
	return rv;

bad:
	pccshw_slot_shutdown(sid);
	return rv;
}

int
pccs_card_open(sid, ci)
	slot_handle_t sid;
	struct card_info *ci;
{
	struct slot_device_iomem *wp;
	struct card_registers *cr;
	u_int8_t index;
	u_long ofs;
	int attr, i, id, zone;

	cr = &ci->ci_cr;
	if (cr->cr_offset == 0)
		return 0;

	attr = 1;
	ofs = CIS_COMM_ADDR(cr->cr_offset);
	index = PCCS_GET_INDEX(ci);
	if ((index & CCOR_IDXMASK) == 0)
		index |= CCOR_MFEN | CCOR_MIEN;
	if (pccs16_map_mem(sid, ofs, attr) != 0)
		return EINVAL;

	zone = 0;
	cr->cr_iob = 0;
	for (i = 0; i < SLOT_DEVICE_NIO; i ++)
	{
		if (ci->ci_res.dr_io[i].im_base != PUNKADDR)
		{
	    		zone |= (ci->ci_res.dr_io[i].im_flags & WFIO_ZONE);
			cr->cr_iob = ci->ci_res.dr_io[i].im_base;
			break;
		}
	}

	cr->cr_iol = 0;
	for (i = 0; i < SLOT_DEVICE_NIO; i ++)
	{
		if (ci->ci_res.dr_io[i].im_size != PUNKSZ)
			cr->cr_iol += ci->ci_res.dr_io[i].im_size;
	}


	pccs16_cr_write_1(ofs, cisr_ccor, CCOR_RST);
	pccs_delay(ci->ci_delay);
	pccs16_cr_write_1(ofs, cisr_ccor, 0);
	pccs_delay(ci->ci_delay);

	if ((cr->cr_mask & BCISR_IOB0) != 0 && zone != 0 &&
	     cr->cr_iob != 0 && cr->cr_iol != 0)
		index |= CCOR_MIOBEN;
	index |= cr->cr_ccmor & CCOR_MMASK;
	index &= ~CCOR_INTR;
	pccs16_cr_write_1(ofs, cisr_ccor, index);
	pccs_delay(1000);

	if (cr->cr_mask & BCISR_CCSR)
	{
		u_int8_t ccsr = cr->cr_ccsr;

		pccs16_cr_write_1(ofs, cisr_ccsr, ccsr);
	}

	for (i = 0; i < sizeof(cr->cr_iob); i++)
	{
		if ((cr->cr_mask & (BCISR_IOB0 << i)) == 0)
			continue;

		/* little endian machine! */
		pccs16_cr_write_1(ofs, cisr_iob0 + 2 * i,
				  cr->cr_iob >> (NBBY * i));
	}

	/* check if I/O Limit register exists */
	if ((cr->cr_mask & BCISR_IOL0) != 0)
	{
		if (cr->cr_iol > 0)
			cr->cr_iol --;
		pccs16_cr_write_1(ofs, cisr_iol, cr->cr_iol);
	}

	pccs_delay(100);

	/* fix up window */
	for (id = i = 0; i < SLOT_DEVICE_NIO; i ++)
	{
		wp = &ci->ci_res.dr_im[i];
		if (wp->im_hwbase == SLOT_DEVICE_UNKVAL)
			continue;
		if (wp->im_type != SLOT_DEVICE_SPIO)
			continue;
		if (ci->ci_iobits < 16)			/* XXX */
			wp->im_hwbase |= 0x4000;	/* XXX */
		pccshw_iomem_map(sid, id ++, wp);
	}
	return 0;
}

int
pccs_close(sid, ci)
	slot_handle_t sid;
	struct card_info *ci;
{

	cisbufp = NULL;
	pccshw_slot_shutdown(sid);
	return 0;
}

/********************************************
 * sub functions
 *******************************************/
#undef	PCCSHW_TRUNC_PAGE
#define	PCCSHW_TRUNC_PAGE(addr) (CIS_COMM_ADDR((addr)) & (~(NBPG - 1)))

static __inline void
pccs16_cr_write_1(ofs, reg, val)
	u_long ofs;
	u_int reg;
	u_int8_t val;
{
	u_int8_t *tp = (u_int8_t *) PCCS_PHYSTOV(attrwin.im_hwbase);
	bus_addr_t addr = CIS_COMM_ADDR(ofs + reg);

	addr = addr - PCCSHW_TRUNC_PAGE(addr);
	tp[addr] = val;
}

static int
pccs16_map_mem(sid, ofs, attr)
	slot_handle_t sid;
	bus_addr_t ofs;
	int attr;
{
	struct slot_device_iomem win;
	int error;

	win = attrwin;
	if (attr != 0)
		win.im_flags |= WFMEM_ATTR;
	else
		win.im_flags &= ~WFMEM_ATTR;
	win.im_base = PCCSHW_TRUNC_PAGE(ofs);
	win.im_size = NBPG;

	error  = pccshw_iomem_map(sid, SLOT_DEVICE_NIO, &win);
	pccs_delay(100);					/* XXX */
	return error;	
}

/********************************************
 * includes
 *******************************************/
#define	CARDINFO
#include "i386/pccs/pccs.c"
#include "i386/pccs/pccscis.c"

#endif	/* BOOT_PCCS */
