/*	$NecBSD: pcio.c,v 1.3.4.1 2000/06/20 06:30:32 honda Exp $	*/
/*	$NetBSD: pcio.c,v 1.8 1999/03/12 04:14:37 sommerfe Exp $	 */

#ifndef	ORIGINAL_CODE
/*
 * [NetBSD for NEC PC-98 series]
 *  Copyright (c) 1998, 1999, 2000
 *	NetBSD/pc98 porting staff. All rights reserved.
 */
#endif	/* PC-98 */
/*
 * Copyright (c) 1996, 1997
 *	Matthias Drochner.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed for the NetBSD Project
 *	by Matthias Drochner.
 * 4. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

/*
 * console I/O
 * needs lowlevel routines from conio.S and comio.S
 */

#include <lib/libsa/stand.h>
#include <lib/libkern/libkern.h>

#include "libi386.h"
#include "bootinfo.h"
#include "comio_direct.h"

#define POLL_FREQ 	10
#define	COM0_IOBASE	0x30

struct btinfo_console btinfo_console;

static void *std_init __P((struct pf_attach_args *));
static int std_putc __P((void *, int));
static int std_getc __P((void *));
static int std_status __P((void *));

struct pcio_funcs {
	void *(*pf_init) __P((struct pf_attach_args *));
	int (*pf_putc) __P((void *, int));
	int (*pf_getc) __P((void *));
	int (*pf_status) __P((void *));
};

struct pcio_funcs pcio_funcs[] = {
	{
		std_init,
		std_putc,
		std_getc,
		std_status,
	},
#ifdef	SUPPORT_SERIAL
	{
		cominit_d,
		computc_d,
		comgetc_d,
		comstatus_d
	}
#endif	/* SUPPORT_SERIAL */
};

static struct pcio_funcs *stdp = &pcio_funcs[0];
static struct pcio_funcs *pfp = &pcio_funcs[0];
static void *pf_arg, *std_arg;

static void *
std_init(pa)
	struct pf_attach_args *pa;
{

	return NULL;
}

static int
std_putc(arg, c)
	void *arg;
	int c;
{
	extern void conputc __P((int));

	conputc(c);
	return 0;
}

static int
std_getc(arg)
	void *arg;
{
	extern int congetc __P((void));

	return (congetc());
}

static int
std_status(arg)
	void *arg;
{
	extern int coniskey __P((void));

	return (coniskey());
}

#ifdef SUPPORT_SERIAL
void
initio(dev)
	int             dev;
{
	struct pf_attach_args pa;

	switch (dev) 
	{
	    case CONSDEV_AUTO:
	    case CONSDEV_COM0:
		(*stdp->pf_putc) (std_arg, '0');

		pa.pa_iobase = COM0_IOBASE;
		pa.pa_speed = CONSPEED;

		pfp = &pcio_funcs[1];
		pf_arg = ((*pfp->pf_init) (&pa));
		if (pf_arg == NULL)
			goto std;

		if ((dev == CONSDEV_COM0) ||
		    (((*pfp->pf_putc)(pf_arg, ':')) &&
		     ((*pfp->pf_putc)(pf_arg, '-')) &&
		     ((*pfp->pf_putc)(pf_arg, '('))))
		{
			btinfo_console.addr = pa.pa_iobase;
			btinfo_console.speed = pa.pa_speed;
			strncpy(btinfo_console.devname, "com", 16);
			(*stdp->pf_putc) (std_arg, '\015');
			(*stdp->pf_putc) (std_arg, '\n');
			break;
		}
		goto std;

	    case CONSDEV_COM1:
	    case CONSDEV_COM2:
	    case CONSDEV_COM3:
	    default:
std:
		pfp = stdp;
		pf_arg = std_arg;
		btinfo_console.speed = 9600;
		strncpy(btinfo_console.devname, "pc", 16);
		break;
	}
}
#else /* !SUPPORT_SERIAL */
void
initio(dev)
	int             dev;
{

	pfp = stdp;
	pf_arg = std_arg;
	btinfo_console.speed = 9600;
	strncpy(btinfo_console.devname, "pc", 16);
}
#endif /* !SUPPORT_SERIAL */

void
putchar(c)
	int             c;
{

	if (c == '\n')
		(*pfp->pf_putc) (pf_arg, '\r');
	(*pfp->pf_putc) (pf_arg, c);
}

int
getchar()
{

	return ((*pfp->pf_getc) (pf_arg));
}

int
iskey()
{

	return ((*pfp->pf_status) (pf_arg));
}

char
awaitkey(timeout, tell)
	int timeout, tell;
{
	int i;
	char c = 0;

	i = timeout * POLL_FREQ;

	while (i) {
		if (tell && (i % POLL_FREQ) == 0)
			printf("%d\b", i/POLL_FREQ);
		if (iskey()) {
			/* flush input buffer */
			while (iskey())
				c = getchar();
			goto out; /* XXX what happens if c == 0? */
		}
		delay(1000000 / POLL_FREQ);
		i--;
	}

out:
	if (tell)
		printf("0\n");

	return(c);
}
