/*	$NecBSD: diskgen.c,v 1.8.4.5.2.7 2000/07/24 02:40:44 honda Exp $	*/
/*	$NetBSD$	*/

/*
 * [NetBSD for NEC PC-98 series]
 *  Copyright (c) 1996, 1997, 1998
 *	NetBSD/pc98 porting staff. All rights reserved.
 * 
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. The name of the author may not be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#define	FSTYPENAMES
#define	DKTYPENAMES
#include <stdio.h>
#include <sys/param.h>
#include <stdlib.h>
#include <unistd.h>
#include <disktab.h>
#include <sys/mount.h>
#include <ufs/ufs/ufsmount.h>
#include <sys/stat.h>
#include <sys/disklabel.h>
#include <sys/fcntl.h>
#include <sys/ioctl.h>
#include <sys/signal.h>
#include <sys/dkio.h>
#include <errno.h>
#include <curses.h>
#include <string.h>
#include <machine/diskutil.h>
#include <ufs/ufs/dinode.h>
#include <ufs/ffs/fs.h>

#define	MAXDOSPART	8

/*******************************************************************
 * declare
 *******************************************************************/
void display __P((FILE *f, struct disklabel *));
int format_request __P((u_char *));
int setup __P((struct disklabel *, u_int, u_int)); 
u_char *getresp __P((u_char *, int));
int writelabel __P((u_char *, struct disklabel *));
int init_bsdpart __P((int, struct disklabel *, struct dos_partition *));
int newfs __P((u_char *, struct disklabel *));
int install_boot __P((u_char *, struct disklabel *));
void get_avail __P((struct disklabel *, u_int *, u_int *));
void clear_partition __P((struct disklabel *, int));
void help __P((void));

void WinInit __P((void));
void WinClose __P((void));
void WinDisplayWarn __P((char *));
void WinSwichW __P((u_char *, struct disklabel *));
void WinDisplayLabelTop __P((u_char *, struct disklabel *));
void WinDisplayPartition __P((struct disklabel *));
void Redraw __P((u_char *, struct disklabel *));
void WinDisplayDosPartition __P((u_char *, struct disklabel *, struct dos_partition *, int));
u_int convert __P((u_char *));
void dos_help __P((void));
int dos_request __P((u_char *));
int boot_request __P((u_char *));
void WinSwichRaw __P((u_char *, char *));
void WinDisplayBootflags __P((u_char *, struct disklabel *, u_int32_t *));
void SigExit __P((int));
int install_dos_bootselect __P((struct disklabel *, u_char *));

int main __P((int, u_char **));

/*******************************************************************
 * var
 *******************************************************************/
u_char *unitchar;
int unitcyl = 0;
int docmd = 1;
int dosslot = -1;
u_char *version = "1.11";
u_int reserve_sec, start_cyl, last_cyl;

/*******************************************************************
 * Curses operations
 *******************************************************************/
WINDOW *pwin, *topwin, *cmdwin, *ewin;

void
WinInit(void)
{

	initscr();
	refresh();
	topwin = subwin(stdscr, 7, COLS - 1, 0, 0);
	pwin = subwin(stdscr, 13, COLS - 1, 7 - 1, 0);
	cmdwin = subwin(stdscr, 3, COLS - 1, 6 + 13, 0);
	ewin = subwin(stdscr, 1, COLS - 1, LINES - 2, 0);
}

void
WinClose(void)
{

	clear();
	refresh();
	endwin();
}

void
SigExit(sig)
	int sig;
{
	
	WinClose();
	exit(1);
}

void
WinDisplayLabelTop(diskname, lp)
	u_char *diskname;
	struct disklabel *lp;
{
	int row, col;

	col = 3;
	row = 1;

	wclear(topwin);
	box(topwin, '*', '-');

	wmove(topwin, row ++, col);
	if (dosslot < 0)
		wprintw(topwin, "NO DOS partitions found!");
	else
		wprintw(topwin, 
			"Target dos cyl region %u - %u \treserve %u sectors",
			 start_cyl, last_cyl, reserve_sec);

	row ++;
	wmove(topwin, row ++, col);
	wprintw(topwin, "device:%s   type:%s  disk: %.*s  label: %.*s", 
		diskname, 
		dktypenames[lp->d_type],
		sizeof(lp->d_typename), lp->d_typename,
		sizeof(lp->d_packname), lp->d_packname);

	wmove(topwin, row ++, col);
	wprintw(topwin, 
		"bytes/sector:%ld   sectors/track:%ld   tracks/cylinder:%ld", 
		(long) lp->d_secsize, 
		(long) lp->d_nsectors,
		(long) lp->d_ntracks);

	wmove(topwin, row ++, col);
	wprintw(topwin, 
		"sectors/cylinder:%ld   cylinders:%ld   total sectors:%ld", 
		(long) lp->d_secpercyl,
		(long) lp->d_ncylinders,
		(long) lp->d_secperunit);

	wrefresh(topwin);
}

void
WinDisplayPartition(lp)
	struct disklabel *lp;
{
	int i, row, col;
	double total, used;
	u_int totali, usedi;
 	struct partition *pp = lp->d_partitions;
	
	col = row = 1;
	wclear(pwin);
	box(pwin, '*', '-');

	wmove(pwin, row ++, col);
	wprintw(pwin, "         size   offset    fstype   [fsize bsize cpg]");

	for (i = 0; i < MAXPARTITIONS; i++, pp++)
	{
		wmove(pwin, row, col);
		if (pp->p_size) 
		{
			double size;

			wprintw(pwin, "  %c: %8d %8d  ", 'a' + i,
				pp->p_size, pp->p_offset);
			if ((unsigned) pp->p_fstype < FSMAXTYPES)
				wprintw(pwin, 
					"%8.8s", fstypenames[pp->p_fstype]);
			else
				wprintw(pwin, "%8d", pp->p_fstype);

			switch (pp->p_fstype)
			{
			case FS_UNUSED:				/* XXX */
				wprintw(pwin, "    %5d %5d %5.5s ",
					pp->p_fsize, 
					pp->p_fsize * pp->p_frag, "");
				break;

			case FS_BSDFFS:
				wprintw(pwin, "    %5d %5d %3d ",
				    pp->p_fsize, pp->p_fsize * pp->p_frag,
				    pp->p_cpg);
				break;

			default:
				wprintw(pwin, "%20.20s", "");
				break;
			}

			wmove(pwin, row, col + 51);
			size = pp->p_size;
			size = size / (1024 * 1024 / lp->d_secsize);
			wprintw(pwin, " %6.1fMB", (float)size);
			wprintw(pwin, " (%5d",
				pp->p_offset / lp->d_secpercyl);
			if (pp->p_offset % lp->d_secpercyl)
				waddch(pwin, '*');
			else
				waddch(pwin, ' ');
			wprintw(pwin, "- %5d", (pp->p_offset + 
				pp->p_size + lp->d_secpercyl - 1) /
				lp->d_secpercyl - 1);
			if (pp->p_size % lp->d_secpercyl)
				waddch(pwin, '*');
			wprintw(pwin, ")");
		}
		else
			wprintw(pwin, 
				"  %c:            --- not defined ---  ", 
				'a' + i);
		row ++;
	}

	row ++;
	get_avail(lp, &totali, &usedi);

	total = totali;
	used = usedi;
	if (unitcyl == 0)
	{
		total = total * lp->d_secpercyl / (1024 * 1024 / lp->d_secsize);
		used = used * lp->d_secpercyl / (1024 * 1024 / lp->d_secsize);
	}

	wmove(pwin, row, col + 2);
	wprintw(pwin, "total %.2f%s used %.2f%s avail %.2f%s",
		(float) total, unitchar, 
		(float) used, unitchar, 
		(float) (total - used), unitchar);

	wrefresh(pwin);
}

void
Redraw(diskname, lp)
	u_char *diskname;
	struct disklabel *lp;
{

	clear();
	refresh();
	WinDisplayLabelTop(diskname, lp);
	WinDisplayPartition(lp);
}

void
WinDisplayWarn(msg)
	char *msg;
{
	
	wmove(ewin, 0, 0);
	wclrtoeol(ewin);
	wprintw(ewin, "Error: %s", msg);
	wrefresh(ewin);
}	

u_char *
getresp(msg, offset)
	u_char *msg;
	int offset;
{
#define	BUFSIZE	1024
	static u_char reqbuf[BUFSIZE];
	int col, c, i, len = 0;
	
	col = 0;
	wmove(cmdwin, offset, col);
	if (offset == 0)
		wclear(cmdwin);
	else
		wclrtoeol(cmdwin);

	if (msg != NULL)
	{
		wprintw(cmdwin, "%s ", msg);
		len = strlen(msg);
	}
	wmove(cmdwin, offset, col + len + 1);

	wrefresh(cmdwin);

	fpurge(stdin);
	cbreak();
	noecho();
	for (i = 0; i < BUFSIZE - 1; )
	{
		c = wgetch(cmdwin);
		if (c == EOF || c == '\n' || c == ' ')
			break;
		if (c == '\b' || c == 0177)
		{
			if (i > 0)
			{
				i --;
				waddch(cmdwin, '\b');
				wdelch(cmdwin);
			}
		}
		else if (c >= 0x20)
		{
			reqbuf[i ++] = c;
			waddch(cmdwin, c);
		}
		wrefresh(cmdwin);
	}
	echo();
	nocbreak();
	reqbuf[i] = 0;
	return reqbuf;
}

void
WinSwichRaw(diskname, msg)
	u_char *diskname;
	char *msg;
{
	
	clear();
	move(0,0);
	refresh();
	if (msg != NULL)
	{
		printf("%s\n", msg);
		fflush(stdout);
	}
}

void
WinSwichW(diskname, lp)
	u_char *diskname;
	struct disklabel *lp;
{

	printf("PUSH RETURN to come back screen:");
	fflush(stdout);
	getchar();

	Redraw(diskname, lp);
}

/*******************************************************************
 * main
 *******************************************************************/
int
main(argc, argv)
	int argc;
	u_char **argv;
{
	extern int opterr, optind;
	extern char *optarg;
	int rv, ch, dosedit = 0, bootflags = 0;
	u_char diskname[FILENAME_MAX];

	if (getenv("TERM") == NULL)
	{
		printf("The TERM variable not specified.\n");
		printf("Please specify the TERM enviorment variable!\n");
		exit(1);
	}

	if (isatty(0) == 0)
	{
		printf("This program require a console device\n");
		exit(1);
	}

	opterr = 0;
	while ((ch = getopt(argc, (char * const *) argv, "sbv")) != EOF)
	{

		switch (ch)
		{
		case 's':
			dosedit = 1;
			break;
		case 'b':
			bootflags = 1;
			break;
		case 'v':
			printf("%s: version %s\n", argv[0], version);
			exit(0);
		}
		opterr = 0;
	}
	
	argc -= optind;
	argv += optind;

	if (argc != 1)
	{
		printf("Specify diskname (wd?, sd?, fd?, vnd?, etc.)\n");
		exit(1);
	}

	unitchar = "MB";
	strcpy(diskname, argv[0]);
	signal(SIGINT, SigExit);
	signal(SIGHUP, SigExit);

	WinInit();

	if (bootflags == 1)
	{
		rv = boot_request(diskname);
	}
	else if (dosedit == 1)
	{
		rv = dos_request(diskname);
	}
	else
	{
		rv = format_request(diskname);
	}

	if (rv != 0)
	{
		sleep(3);
	}
	WinClose();
	return rv;
}

/*******************************************************************
 * Format Request
 *******************************************************************/
void
help(void)
{
	clear();
	box(stdscr, '|', '-');
	move(1,3);printw("<HELP>");
	move(3,3); printw("p or print -- show current settings.");
	move(4,3); printw("label or write -- write current disklabel and newfs.");
	move(5,3); printw("q or quit  -- exit the program.");
	move(6,3); printw("u or unit  -- change units MB <=> cyls (toggle).");
	move(7,3); printw("partition name (a, b, e, f, g, h) -- set partition size.");
	move(8,3); printw("getdiskbyname -- get entry data in disktab.");
	move(9,3); printw("restore    -- restore old disklabel data.");
	move(10,3);printw("select     -- select target disk.");
	move(LINES - 2,1); printw("push return:");
	refresh();
	getchar();
}

int
format_request(diskname)
	u_char *diskname;
{
	struct disklabel mydisklabel, reqdisklabel;
	struct disklabel *lp;
	u_char *ans;
	int fd, part;
	u_char buf[BBSIZE];
	u_char devname[FILENAME_MAX];
	u_char cmd[FILENAME_MAX];
	u_long ecyl;

Loop:
	lp = &mydisklabel;
	bzero(lp, sizeof(*lp));

	sprintf(devname, "/dev/r%sd", diskname);
	fd = open(devname, O_RDONLY, 0);
	if (fd < 0)
	{
		wprintw(ewin, "Error: can not open %s", devname);
		wrefresh(ewin);
		return EINVAL;
	}

	if (ioctl(fd, DIOCGDINFO, lp) < 0)
	{
		wprintw(ewin, "Error: DIOCGDINFO failed");
		close(fd);
		wrefresh(ewin);
		return EIO;
	}
	close(fd);

	if (strncmp(diskname, "wd", 2) == 0)
		reserve_sec = lp->d_secpercyl * 2;
	else
		reserve_sec = 0;

	if (disk_netbsd_start_sector(lp) == disk_total_start_sector(lp))
		dosslot = -1;
	else
		dosslot = 0;

	start_cyl = disk_netbsd_start_sector(lp) / lp->d_secpercyl;
	ecyl = (disk_netbsd_start_sector(lp) + disk_netbsd_size_sector(lp)) / lp->d_secpercyl - 1;

	if (ecyl < start_cyl || ecyl > lp->d_ncylinders - 1)
	{
		WinDisplayWarn("target partitions corrupted");
		return EINVAL;
	}

	last_cyl = ecyl;
	unitchar = "MB";

	/*
	 * setup c parition.
	 */
	reqdisklabel = mydisklabel;

 	lp = &reqdisklabel;
	lp->d_partitions[RAW_PART - 1].p_offset = start_cyl * lp->d_secpercyl;
	lp->d_partitions[RAW_PART - 1].p_size = (last_cyl - start_cyl + 1) * lp->d_secpercyl;

	/*
	 * initialize all user's partitions.
	 */
	lp->d_npartitions = RAW_PART + 1;
	for (part = 0; part < MAXPARTITIONS; part ++)
	{
		if (part == RAW_PART - 1 || part == RAW_PART)
			continue;
		clear_partition(lp, part);
	}

	Redraw(diskname, lp);

	while(1)
	{
		WinDisplayPartition(lp);

		ans = getresp("what>", 0);
		if (strcmp(ans, "quit") == 0 || strcmp(ans, "exit") == 0 || 
		    *ans == 'q' || *ans == 'x')
			break;

		if (strcmp(ans, "print") == 0 || *ans == 'p')
		{
			WinSwichRaw(diskname, NULL);
			display(stdout, lp);
			WinSwichW(diskname, lp);
			continue;
		}

		if (strcmp(ans, "unit") == 0 || *ans == 'u')
		{
			unitcyl ^= 1;
			if (unitcyl)
				unitchar = "cyls";
			else
				unitchar = "MB";
			continue;
		}

		if (strcmp(ans, "restore") == 0)
		{
			*lp = mydisklabel;
			Redraw(diskname, lp);
			continue;
		}

		if (strcmp(ans, "select") == 0)
		{
			ans = getresp("target disk name:", 1);
			if (*ans == 0)
				continue;

			sprintf(devname, "/dev/r%sd", ans);
			fd = open(devname, O_RDWR, 0);
			if (fd < 0)
			{
				WinDisplayWarn( "can not open device");
				continue;
			}
			close(fd);
			strcpy(diskname, ans);
			goto Loop;
		}
			
		if (strcmp(ans, "getdiskbyname") == 0)
		{
			struct disklabel *tlp;

			ans = getresp("disktype name:", 1);
			if (*ans == 0)
				continue;

			tlp = getdiskbyname(ans);
			if (tlp == NULL)
			{
				WinDisplayWarn( "can not find entry in disktab");
				continue;
			
			}

			*lp = *tlp;
			Redraw(diskname, lp);
			continue;
		}

		if (strcmp(ans, "label") == 0 || strcmp(ans, "write") == 0)
		{
			ans = getresp("disk name:", 1);
			if (*ans != 0)
				strncpy(lp->d_typename, ans,
					sizeof(lp->d_typename) - 1);	
			ans = getresp("label name:", 1);
			if (*ans != 0)
				strncpy(lp->d_packname, ans,
				        sizeof(lp->d_packname) - 1);	
			if (writelabel(diskname, lp) != 0)
				continue;
			if (newfs(diskname, lp) != 0)
				continue;
			if (install_boot(diskname, lp) != 0)
				continue;
			ans = getresp("change bootflags (y/n)? ", 2);
			if (*ans == 'y')
			{
				WinSwichRaw(diskname, "change bootflags  ....");
				sprintf(cmd, "/sbin/diskgen -b %s", diskname);
				system(cmd);
			}
			Redraw(diskname, lp);
			continue;
	 	}		

		if (ans[1] == 0 && *ans >= 'a' && *ans <= 'h')
		{
			u_char c = *ans;
			u_int part = c - 'a';
			double size;

			if (part == RAW_PART - 1 || part == RAW_PART)
			{
				WinDisplayWarn( "can not change c or d partitions");
				continue;
			}

			ans = getresp("partition size (* means all avail):", 1);
			if (*ans == 0)
				continue;

			if (*ans == '*')
			{
				u_int total, used;

				clear_partition(lp, part);
				get_avail(lp, &total, &used);
				size = (total - used) * lp->d_secpercyl;
			}
			else
			{
				size = atof(ans);
				if (unitcyl)
					size *= lp->d_secpercyl;
				else
					size *= (1024 * 1024 / lp->d_secsize);
			}
			setup(lp, part, (u_int) size);
			continue;
		}
		
		if (*ans == 0)
		{
			Redraw(diskname, lp);
			continue;
		}

		if (strcmp(ans, "help") == 0 || *ans == '?')
		{
			help();
			Redraw(diskname, lp);
		}
	}
	return 0;
}
			
void
clear_partition(lp, part)
	struct disklabel *lp;
	int part;
{
	struct partition *pp;
	
	pp = &lp->d_partitions[part];
	pp->p_size = 0;
	pp->p_fstype = FS_UNUSED;
	pp->p_offset = 0;
	pp->p_fsize = 0;
	pp->p_frag = 0;
	pp->p_cpg = 0;
}

void
get_avail(lp, totalp, usedp)
	struct disklabel *lp;
	u_int *totalp, *usedp;
{
	u_int ssec, part;
	
	for (ssec = part = 0; part < MAXPARTITIONS; part ++)
	{
		if (part == RAW_PART - 1 || part == RAW_PART)
			continue;
		
		if (lp->d_partitions[part].p_offset <
		    lp->d_partitions[RAW_PART - 1].p_offset ||
		    lp->d_partitions[part].p_offset >=
		    lp->d_partitions[RAW_PART - 1].p_offset +
		    lp->d_partitions[RAW_PART - 1].p_size)
			continue;
		ssec += lp->d_partitions[part].p_size;
	}

	*totalp = lp->d_partitions[RAW_PART - 1].p_size / lp->d_secpercyl;
	*usedp = (ssec + reserve_sec) / lp->d_secpercyl;
}

int
setup(tlp, tpart, secs)
	struct disklabel *tlp;
	u_int tpart, secs;
{
	u_int part, lastpart, ssec;
	struct disklabel tmplabel;
	struct disklabel *lp = &tmplabel;
	struct partition *pp, *ppc;

	if (tpart == RAW_PART - 1 || tpart == RAW_PART)
		return 1;

	*lp = *tlp;
	secs -= secs % lp->d_secpercyl;
	lp->d_partitions[tpart].p_size = secs;

	ppc = &lp->d_partitions[RAW_PART - 1];
	for (ssec = part = lastpart = 0; part < MAXPARTITIONS; part ++)
	{
		if (part == RAW_PART - 1 || part == RAW_PART)
			continue;
		
		pp = &lp->d_partitions[part];
		if (pp->p_size == 0)
		{
			clear_partition(lp, part);
			continue;
		}
		if (tpart != part &&
		    (pp->p_offset < ppc->p_offset ||
		     pp->p_offset >= ppc->p_offset + ppc->p_size))
		{
			clear_partition(lp, part);
			continue;
		}
		
		pp->p_offset = ssec + ppc->p_offset;
		if (ssec + reserve_sec + pp->p_size >= ppc->p_size)
			pp->p_size = ppc->p_size - ssec - reserve_sec;

		ssec += pp->p_size;

		if (part == 1)
		{
			pp->p_fstype = FS_SWAP;
		}
		else
		{
			pp->p_fstype = FS_BSDFFS;
			pp->p_fsize = 1024;
			pp->p_frag = 8;
			pp->p_cpg = 16;
		}
		lastpart = part;
	}

	for ( ; part < MAXPARTITIONS; part ++)
	{
		if (part == RAW_PART - 1 || part == RAW_PART)
			continue;
		clear_partition(lp, part);
	}

	if (lastpart < RAW_PART)
		lp->d_npartitions = RAW_PART + 1;
	else		
		lp->d_npartitions = lastpart + 1;

	*tlp = *lp;
	return 0;
}

void
display(f, lp)
	FILE *f;
	struct disklabel *lp;
{
	int i, j;
	struct partition *pp;

	if ((unsigned) lp->d_type < DKMAXTYPES)
		fprintf(f, "type: %s\n", dktypenames[lp->d_type]);
	else
		fprintf(f, "type: %d\n", lp->d_type);
	fprintf(f, "disk: %.*s\n", (int) sizeof(lp->d_typename), lp->d_typename);
	fprintf(f, "label: %.*s\n", (int) sizeof(lp->d_packname), lp->d_packname);
	fprintf(f, "flags:");
	if (lp->d_flags & D_REMOVABLE)
		fprintf(f, " removable");
	if (lp->d_flags & D_ECC)
		fprintf(f, " ecc");
	if (lp->d_flags & D_BADSECT)
		fprintf(f, " badsect");
	fprintf(f, "\n");
	fprintf(f, "bytes/sector: %ld\n", (long) lp->d_secsize);
	fprintf(f, "sectors/track: %ld\n", (long) lp->d_nsectors);
	fprintf(f, "tracks/cylinder: %ld\n", (long) lp->d_ntracks);
	fprintf(f, "sectors/cylinder: %ld\n", (long) lp->d_secpercyl);
	fprintf(f, "cylinders: %ld\n", (long) lp->d_ncylinders);
	fprintf(f, "total sectors: %ld\n", (long) lp->d_secperunit);
	fprintf(f, "rpm: %ld\n", (long) lp->d_rpm);
	fprintf(f, "interleave: %ld\n", (long) lp->d_interleave);
	fprintf(f, "trackskew: %ld\n", (long) lp->d_trackskew);
	fprintf(f, "cylinderskew: %ld\n", (long) lp->d_cylskew);
	fprintf(f, "headswitch: %ld\t\t# milliseconds\n",
		(long) lp->d_headswitch);
	fprintf(f, "track-to-track seek: %ld\t# milliseconds\n",
		(long) lp->d_trkseek);
	fprintf(f, "drivedata: ");
	for (i = NDDATA - 1; i >= 0; i--)
		if (lp->d_drivedata[i])
			break;
	if (i < 0)
		i = 0;
	for (j = 0; j <= i; j++)
		fprintf(f, "%d ", lp->d_drivedata[j]);
	fprintf(f, "\n\n%d partitions:\n", lp->d_npartitions);

	fprintf(f,
	    "#        size   offset    fstype   [fsize bsize   cpg]\n");
	pp = lp->d_partitions;
	for (i = 0; i < lp->d_npartitions; i++, pp++) {
		if (pp->p_size) {
			fprintf(f, "  %c: %8d %8d  ", 'a' + i,
			   pp->p_size, pp->p_offset);
			if ((unsigned) pp->p_fstype < FSMAXTYPES)
				fprintf(f, "%8.8s", fstypenames[pp->p_fstype]);
			else
				fprintf(f, "%8d", pp->p_fstype);
			switch (pp->p_fstype) {

			case FS_UNUSED:				/* XXX */
				fprintf(f, "    %5d %5d %5.5s ",
				    pp->p_fsize, pp->p_fsize * pp->p_frag, "");
				break;

			case FS_BSDFFS:
				fprintf(f, "    %5d %5d %5d ",
				    pp->p_fsize, pp->p_fsize * pp->p_frag,
				    pp->p_cpg);
				break;

			default:
				fprintf(f, "%20.20s", "");
				break;
			}
			fprintf(f, "\t# (Cyl. %4d",
			    pp->p_offset / lp->d_secpercyl);
			if (pp->p_offset % lp->d_secpercyl)
			    putc('*', f);
			else
			    putc(' ', f);
			fprintf(f, "- %d",
			    (pp->p_offset + 
			    pp->p_size + lp->d_secpercyl - 1) /
			    lp->d_secpercyl - 1);
			if (pp->p_size % lp->d_secpercyl)
			    putc('*', f);
			fprintf(f, ")\n");
		}
	}
	fflush(f);
}

int
init_bsdpart(fd, lp, tdp)
	int fd;
	struct disklabel *lp;
	struct dos_partition *tdp;
{
	struct disklabel *nlp;
	u_long sector;
	off_t offset;
	u_int8_t bootbuf[BBSIZE];

	sector = disk_start_sector(lp, tdp);
	offset = (off_t) sector * (off_t) lp->d_secsize;
	if (lseek(fd, offset, 0) != offset)
		return EIO;

	read(fd, bootbuf, BBSIZE);
	nlp = (struct disklabel *) (bootbuf + lp->d_secsize * LABELSECTOR);
	if (nlp->d_magic != DISKMAGIC)
		return 0; 

	nlp->d_magic = 0;
	nlp->d_magic2 = 0;
	if (lseek(fd, offset, 0) != offset)
		return EIO;

	write(fd, bootbuf, BBSIZE);
	return 0;
}
	
int
writelabel(diskname, lp)
	u_char *diskname;
	struct disklabel *lp;
{
	u_char *ans, *options;
	FILE *f;
	int error;
	u_char tempfile[FILENAME_MAX];
	u_char cmd[FILENAME_MAX];
	
	ans = getresp("write disklabel? (y/n)", 1);
	if (*ans != 'y')
		return 1;
	options = "-r -R";

	strcpy(tempfile, "/tmp/disklabel.XXXX");
	mkstemp(tempfile);
	f = fopen(tempfile, "w+");
	if (f == NULL)
	{
		WinDisplayWarn("can not make temp file");
		return 1;
	}

	display(f, lp);
	fclose(f);

	sprintf(cmd, "/sbin/disklabel %s %s %s\n", options, diskname, tempfile);	
	WinSwichRaw(diskname, "writing disklabel ....");
	if (docmd != 0)
		error = system(cmd);
	else
	{
		printf("%s", cmd);
		error = 0;
	}
	WinSwichW(diskname, lp);

	return error;
}	

static int exec_cmd __P((u_char *));
static int copy_bootflags __P((u_char *, u_char *, int));

static int
exec_cmd(s)
	u_char *s;
{
	int error;

	printf("%s\n", s);
	fflush(stdout);
	return system(s);
}

static int getmountpoint __P((char *, char *));

static int
getmountpoint(diskdev, outbuf)
	char *diskdev;
	char *outbuf;
{
	u_char bdiskdev[FILENAME_MAX];
	struct statfs *buf;
	int num, i;

	sprintf(bdiskdev, "/dev/%sa", diskdev);

	num = getmntinfo(&buf, MNT_WAIT);
	if (num == 0)
		return -1;

	for (i = 0; i < num; i++)
	{
		if (strncmp(bdiskdev, buf[i].f_mntfromname, MNAMELEN) == 0)
		{
			int j;

			for (j = 0; j < num; j++)
				if ((i != j) && (strncmp(buf[i].f_mntonname,
							 buf[j].f_mntonname,
							 MNAMELEN) == 0))
			{
				return (-1);
			}

			/* this code is FFS only */
			if (strncmp(buf[i].f_fstypename, MOUNT_FFS, MFSNAMELEN)) {
				return (-1);
			}
			strncpy(outbuf, buf[i].f_mntonname, FILENAME_MAX);
			return 0;
		}
	}

	sprintf(outbuf, "/tmp/diskgen.%s.mtpt", diskdev);
	return 1;
}

static int
copy_bootflags(diskname, target_file, do_edit)
	u_char *diskname, *target_file;
	int do_edit;
{
	int error, mntop;
	u_char mtpt[FILENAME_MAX];
	u_char cmd[BUFSIZE];
	char *editor;

	editor = getenv("EDITOR");
	if (editor == NULL)
		editor = "/bin/ed";

	mntop = getmountpoint(diskname, mtpt);
	if (mntop < 0)
		return EINVAL;

	if (mntop > 0)
	{
		sprintf(cmd, "/bin/mkdir %s", mtpt);
		error = exec_cmd(cmd);

		sprintf(cmd, "/sbin/mount /dev/%sa %s", diskname, mtpt);
		error = exec_cmd(cmd);
		if (error != 0)
			goto out;
	}

	sprintf(cmd, "/bin/mkdir %s/etc", mtpt);
	error = exec_cmd(cmd);

	if (target_file != NULL)
	{
		sprintf(cmd, "/bin/cp -p "
			     "%s/bootflags.conf %s/etc/bootflags.conf",
			      target_file, mtpt);
		error = exec_cmd(cmd);
	}

	if (do_edit & 0x02)
	{
		sprintf(cmd, "/sbin/bootflags -s -f %s/etc", mtpt);
		error = exec_cmd(cmd);
		if (error != 0)
			goto out;
		sprintf(cmd, "mv %s/etc/bootflags.conf "
			     "%s/etc/bootflags.conf.bak",
			      mtpt, mtpt);
		error = exec_cmd(cmd);

		sprintf(cmd, "/sbin/bootflags -d -f %s/etc "
			     "> %s/etc/bootflags.conf",
			mtpt, mtpt);
		error = exec_cmd(cmd);
		goto out;
	}	

	if (do_edit & 0x01)
	{
		sprintf(cmd, "%s %s/etc/bootflags.conf", editor, mtpt);
		exec_cmd(cmd);
	}

	sprintf(cmd, "/sbin/bootflags -f %s/etc", mtpt);
	error = exec_cmd(cmd);
	if (error != 0)
		goto out;

out:
	if (mntop > 0)
	{
		sprintf(cmd, "/sbin/umount %s", mtpt);
		error = exec_cmd(cmd);
	}
	return error;
}

int
newfs(diskname, lp)
	u_char *diskname;
	struct disklabel *lp;
{
	u_char *ans, cmd[FILENAME_MAX];
	struct partition *pp;
	int part, error = 0;
	
	ans = getresp("newfs now? (y/n)", 1);
	if (*ans != 'y')
		return 1;

	WinSwichRaw(diskname, "making BSD file systems  ....");
	for (part = 0; part < lp->d_npartitions; part ++)
	{
		if (part == RAW_PART - 1 || part == RAW_PART)
			continue;	
		
		pp = &lp->d_partitions[part];
		if (pp->p_fstype != FS_BSDFFS || pp->p_fsize == 0)
			continue;

		sprintf(cmd, "/sbin/newfs /dev/r%s%c\n", diskname, part + 'a');
		if (docmd != 0)
		{
			error = system(cmd);
			if (error != 0)
				break;
		}
		else
			printf("%s", cmd);
	}
	WinSwichW(diskname, lp);
	return error;
}

int
install_boot(diskname, lp)
	u_char *diskname;
	struct disklabel *lp;
{
	u_char *ans, cmd[FILENAME_MAX];
	u_char target_file[BUFSIZE];
	int error = 0;

	ans = getresp("install bootloader? (y/n)", 2);
	if (*ans == 'y')
	{
		ans = getresp("bootloader name? [biosboot.sym]: ", 2);
		if (*ans == 0)
			strcpy(target_file, "biosboot.sym");
		else
			strcpy(target_file, ans);
		sprintf(cmd, "/usr/mdec/installboot -v /usr/mdec/%s /dev/r%sa",
			target_file, diskname);

		WinSwichRaw(diskname, "install new boot loader  ....");
		if (docmd != 0)
			error = system(cmd);
		else
			printf("%s", cmd);
		WinSwichW(diskname, lp);
	}

	if (error != 0)
		return error;

	ans = getresp("install bootflags.conf? (y/n)", 2);
	if (*ans == 'y')
	{
		int do_edit = 0;

		ans = getresp("where is your bootflags.conf? [/etc]: ", 2);
		if (*ans == 0)
			strcpy(target_file, "/etc");
		else
			strcpy(target_file, ans);
		ans = getresp("do you want to edit bootflags.conf (y/n)? ", 2);
		if (*ans == 'y')
			do_edit = 1;
		WinSwichRaw(diskname, "install bootflags.conf  ....");
		error = copy_bootflags(diskname, target_file, do_edit);
		WinSwichW(diskname, lp);
	}
	return error;
}

/*******************************************************************
 * Dos partition Request
 *******************************************************************/
void
WinDisplayDosPartition(diskname, lp, dp, sel)
	u_char *diskname;
	struct disklabel *lp;
	struct dos_partition *dp;
	int sel;
{
	sysid_t id;
	int i, row, col, ocol;
	double size;
	u_char dp_name[18];
	int ndospart;
	
	ndospart = (lp->d_secsize / 32);
	if (ndospart > MAXDOSPART)
		ndospart = MAXDOSPART;

	col = row = 1;
	wclear(pwin);
	box(pwin, '*', '-');

	WinDisplayLabelTop(diskname, lp);
	wmove(pwin, row ++, col);
	wprintw(pwin,
	"      System Name       sid  mid    From    To(cyl)      Size");

	ocol = col;
	for (i = 0; i < ndospart; i++, dp++)
	{
		id = dos_partition_read_sysid(dp);
		if (id == 0)
			break;

		col = ocol;
		row ++;

		wmove(pwin, row, col);
		if (sel == i)
			wprintw(pwin, ">>");
		else
			wprintw(pwin, "  ");

		strncpy(dp_name, dp->dp_name, 16);
		dp_name[16] = 0;
		wprintw(pwin, "%2d  %s", i, dp_name);

		col += 23;
		wmove(pwin, row, col);
		wprintw(pwin, "0x%2x ", (u_int) sysid_to_sid(id));
		wprintw(pwin, "0x%2x ", (u_int) sysid_to_mid(id));

		wprintw(pwin, "%7d   %7d  ", dp->dp_scyl, dp->dp_ecyl);

		size = disk_size_sector(lp, dp);
		size = size / (1024 * 1024 / lp->d_secsize);
		wprintw(pwin, "%7.2fMB  ", size);
	}
	wrefresh(pwin);
}

u_int
convert(s)
	u_char *s;
{
	u_int num, val = 0, radix = 10;

	if (strncmp(s, "0x", 2) == 0)
		radix = 16;

	for (s; *s != 0; s++)
	{
		if (*s >= '0' && *s <= '9')
			num = *s - '0';
		else if (radix == 16 && *s <= 'f' && *s >= 'a')
			num = *s - 'a' + 10;
		else
			continue;

		val = val * radix;
		val += num;
	}

	return val;
}
	
void
dos_help(void)
{
	clear();
	box(stdscr, '|', '-');
	move(1,3);printw("<HELP>");
	move(2,3); printw("q or quit -- exit the program.");
	move(3,3); printw("x or exit -- exit the program.");
	move(4,3); printw("select    -- select target DOS partition.");
	move(5,3); printw("clear     -- clear current setting.");
	move(6,3); printw("NetBSD    -- change target DOS partition to NetBSD/pc98 partition.");
	move(7,3); printw("sid       -- change sid of target DOS partition.");
	move(8,3); printw("mid       -- change mid of target DOS partition.");
	move(9,3); printw("name      -- change system name of target DOS partition.");
	move(10,3);printw("from      -- change a start cylinder of target DOS partition.");
	move(11,3);printw("to        -- change a end cylinder of target DOS partition.");
	move(12,3);printw("delete    -- delete the selected partition.");
	move(13,3);printw("add       -- add a partition.");
	move(14,3);printw("initpart  -- initialize a partition header.");
	move(15,3);printw("initboot  -- initialize boot selector.");
	move(16,3);printw("write     -- write current DOS partition table.");
	move(LINES - 2,1); printw("push return:");
	refresh();
	getchar();
}

int
install_dos_bootselect(lp, bufp)
	struct disklabel *lp;
	u_char *bufp;
{
	int fd;
	int bootsize;
	u_char bootcode[BBSIZE];

	fd = open("/usr/mdec/mbr", O_RDONLY, 0);
	if (fd < 0)
		return EIO;

	bzero(bootcode, BBSIZE);
	bootsize = read(fd, bootcode, BBSIZE);
	close(fd);
	if (bootsize < lp->d_secsize)
		return EIO;

	bcopy(bootcode, bufp, lp->d_secsize);
	if (bootsize > lp->d_secsize * 2)
	{
		bcopy(bootcode + lp->d_secsize * 2,
		      bufp + lp->d_secsize * 2, 
		      bootsize - 2 * lp->d_secsize);
	}
	return 0;
}

int
dos_request(diskname)
	u_char *diskname;
{
	struct dos_partition *dp, *tdp;
	struct disklabel mydisklabel;
	struct disklabel *lp;
	sysid_t id;
	u_char *ans;
	int flag, sel, fd, part, ndospart;
	int install_bootselect, init_diskpart;
	u_char devname[FILENAME_MAX];
	u_char buf[BBSIZE];

	sprintf(devname, "/dev/r%sd", diskname);

again:
	fd = open(devname, O_RDONLY, 0);
	if (fd < 0)
	{
		wprintw(ewin, "Error: can not open %s", devname);
		wrefresh(ewin);
		return EINVAL;
	}

	sel = -1;
	install_bootselect = 0;
	init_diskpart = 0;
	lp = &mydisklabel;
	bzero(lp, sizeof(*lp));

	/*
	 * setup initial disklabel
	 */
	if (ioctl(fd, DIOCGDINFO, lp) < 0)
	{
		wprintw(ewin, "Error: DIOCGDINFO failed");
		goto bad_open;
	}

	if (lseek(fd, (off_t) 0, 0) != (off_t) 0)
	{
		wprintw(ewin, "Error: lseek failed");
		goto bad_open;
	}

	if (read(fd, buf, BBSIZE) != BBSIZE)
	{
		wprintw(ewin, "Error: can not read boot sector");
bad_open:
		close(fd);
		wrefresh(ewin);
		return EIO;
	}
	close(fd);

	ndospart = (lp->d_secsize / 32);
	dp = (struct dos_partition *) (buf + lp->d_secsize);
	tdp = NULL;

	if (disk_netbsd_start_sector(lp) == disk_total_start_sector(lp))
		dosslot = -1;
	else
		dosslot = 0;

	while (1)
	{
		WinDisplayDosPartition(diskname, lp, dp, sel);

		ans = getresp("what>", 0);
		if (strcmp(ans, "quit") == 0 || strcmp(ans, "exit") == 0 || 
		    *ans == 'q' || *ans == 'x')
			break;

		if (*ans == 0)
		{
			Redraw(diskname, lp);
			continue;
		}
		if (strcmp(ans, "help") == 0 || *ans == '?')
		{
			dos_help();
			Redraw(diskname, lp);
			continue;
		}

		if (strcmp(ans, "select") == 0)
		{
			ans = getresp("target partition:", 1);
			if (*ans == 0)
				continue;
			sel = atoi(ans);
			if (sel < 0 || sel >= ndospart)
				WinDisplayWarn( "invalid dos partition");
			tdp = &dp[sel];
			install_bootselect = 0;
			init_diskpart = 0;
			continue;
		}

		if (strcmp(ans, "add") == 0)
		{
			int i;
			
			for (i = 0; i < ndospart; i ++)
			{
				id = dos_partition_read_sysid(&dp[i]);
				if (id == 0)
					break;
			}
			if (i >= ndospart)
				continue;

			memset(&dp[i], 0, sizeof(dp[i]));
			id = sysid_make_id(DOSMID_NETBSD, DOSSID_NETBSD);
			dos_partition_write_sysid(&dp[i], id);
			strncpy(dp[i].dp_name, "NetBSD/pc98", 16);
			continue;
		}

		if (strcmp(ans, "clear") == 0)
			goto again;

		if (tdp == NULL)
		{
			WinDisplayWarn( "select partition please!");
			continue;
		}

		if (strcmp(ans, "delete") == 0)
		{
			int no, i;

			/* packing */
			for(no = sel, i = sel + 1; i < ndospart; i ++, no ++)
				bcopy(dp + i, dp + no, sizeof(*tdp));
			for ( ; no < ndospart; no ++)
				memset(dp + no, 0, sizeof(*tdp));
			sel = -1;
			tdp = NULL;
			continue;
		}

		if (strcmp(ans, "NetBSD") == 0)
		{
			strncpy(tdp->dp_name, "NetBSD/pc98", 16);
			id = sysid_make_id(DOSMID_NETBSD, DOSSID_NETBSD);
			dos_partition_write_sysid(tdp, id);
			continue;
		}

		if (strcmp(ans, "sid") == 0)
		{
			ans = getresp("sid:", 1);
			if (*ans == 0)
				continue;
			
			id = dos_partition_read_sysid(tdp);
			id = sysid_make_id(sysid_to_mid(id), convert(ans));
			dos_partition_write_sysid(tdp, id);
			continue;
		}

		if (strcmp(ans, "mid") == 0)
		{
			ans = getresp("mid:", 1);
			if (*ans == 0)
				continue;
			
			id = dos_partition_read_sysid(tdp);
			id = sysid_make_id(convert(ans), sysid_to_sid(id));
			dos_partition_write_sysid(tdp, id);
			continue;
		}

		if (strcmp(ans, "name") == 0)
		{
			ans = getresp("system name:", 1);
			if (*ans == 0)
				continue;
			strncpy(tdp->dp_name, ans, 16);
			continue;
		}

		if (strcmp(ans, "from") == 0)
		{
			ans = getresp("start cylinder:", 1);
			if (*ans == 0)
				continue;

			tdp->dp_scyl = convert(ans);
			tdp->dp_ipl_cyl = tdp->dp_scyl;
			continue;
		}

		if (strcmp(ans, "to") == 0)
		{
			ans = getresp("end cylinder:", 1);
			if (*ans == 0)
				continue;

			tdp->dp_ecyl = convert(ans);
			continue;
		}

		if (strcmp(ans, "initpart") == 0)
		{
			init_diskpart = 1; 
			continue;
		}

		if (strcmp(ans, "initboot") == 0)
		{
			install_bootselect = 1;
			continue;
		}

		if (strcmp(ans, "write") == 0)
		{
			ans = getresp("really write(y/n):", 1);
			if (*ans == 0 || *ans != 'y')
				continue;

			fd = open(devname, O_RDWR, 0);
			if (fd < 0)
			{
				WinDisplayWarn( "write open failed");
				continue;
			}

			flag = 1;
			if (ioctl(fd, DIOCWLABEL, &flag) < 0)
			{
				WinDisplayWarn( "DIOCWLABEL fail");
				goto bad_write;
			}

			if (init_diskpart != 0)
			{
				ans = getresp("initialize partition(y/n):", 1);
				if (*ans == 'y')
				{
					init_bsdpart(fd, lp, tdp);
				}
			}

			if (dosslot < 0 || install_bootselect > 0)
			{
				ans = getresp("install bootselect(y/n):", 1);
				if (*ans == 'y')
				{
					install_dos_bootselect(lp, buf);
				}
			}

			if (lseek(fd, (off_t) 0, 0) != (off_t) 0)
			{
				WinDisplayWarn( "seek fail");
				goto bad_write;
			}
	
			if (write(fd, buf, BBSIZE) != BBSIZE)
				WinDisplayWarn( "write fail");

bad_write:
			flag = 0;
			ioctl(fd, DIOCWLABEL, &flag);
			close(fd);
			goto again;
		}
	}
	return 0;
}

/*******************************************************************
 * Bootflags Request
 *******************************************************************/
#define	_KERNEL
#include <machine/bootinfo.h>
#include <machine/bootflags.h>
static void __ShowFlags __P((int, int, int, int, u_char *));
void bootflags_help __P((void));

void
bootflags_help(void)
{
	clear();
	box(stdscr, '|', '-');
	move(1,3);printw("<HELP>");
	move(2,3); printw("q or quit -- exit the program.");
	move(3,3); printw("x or exit -- exit the program.");
	move(4,3); printw("0,1,2...  -- toggle the target flag.");
	move(5,3); printw("edit      -- edit current bootflags.conf.");
	move(6,3); printw("update    -- update bootloader.");
	move(7,3); printw("save      -- save current bootflags settings.");
	move(8,3); printw("write     -- write back modifications.");
	move(LINES - 2,1); printw("push return:");
	refresh();
	getchar();
}

static void
__ShowFlags(row, col, nent, flags, msg)
	int row, col, nent, flags;
	u_char *msg;
{
	u_char *s;

	wmove(pwin, row, col);
	wprintw(pwin, "[%d] %s:", nent, msg);
	wmove(pwin, row, col + 40);
	s = ((flags != 0) ? "yes" : "no");
	wprintw(pwin, "%s ", s);
}

void
WinDisplayBootflags(diskname, lp, fp)
	u_char *diskname;
	struct disklabel *lp;
	u_int32_t *fp;
{
	int row, col;
	
	col = row = 1;
	wclear(pwin);
	box(pwin, '*', '-');

	WinDisplayLabelTop(diskname, lp);

	wmove(pwin, row ++, col);

	__ShowFlags(row ++ , col, 0, 
		  *fp & BOOTFLAGS_FLAGS_BIOSGEOM,
		  "get disk geometry");
	__ShowFlags(row ++, col, 1, 
		  *fp & BOOTFLAGS_FLAGS_VERBOSE,
		  "verbose mode");
	__ShowFlags(row ++, col, 2, 
		  *fp & BOOTFLAGS_FLAGS_DEVICES,
		  "device");
	__ShowFlags(row ++, col, 3, 
		  *fp & BOOTFLAGS_FLAGS_DISK_DIRECT,
		  "disk direct read");
	__ShowFlags(row ++, col, 4, 
		  *fp & BOOTFLAGS_FLAGS_CONSOLE_SERIAL,
		  "serial console");
	__ShowFlags(row ++, col, 5, 
		  *fp & BOOTFLAGS_FLAGS_BOOT_NETWORK,
		  "network boot");
	wrefresh(pwin);
}

int
boot_request(diskname)
	u_char *diskname;
{
	struct disklabel mydisklabel;
	struct disklabel *lp;
	u_char devname[FILENAME_MAX];
	u_char buf[BBSIZE];
	struct biosboot_header *bh;
	off_t offset;
	int fd, part, flag;
	char *ans;

	lp = &mydisklabel;
	sprintf(devname, "/dev/r%sd", diskname);

again:
	bzero(lp, sizeof(*lp));
	fd = open(devname, O_RDONLY, 0);
	if (fd < 0)
	{
		wprintw(ewin, "Error: can not open %s", devname);
		wrefresh(ewin);
		return EINVAL;
	}

	/*
	 * setup initial disklabel
	 */
	if (ioctl(fd, DIOCGDINFO, lp) < 0)
	{
		wprintw(ewin, "Error: DIOCGDINFO failed");
		goto bad_open;
	}

	offset = ((off_t) disk_netbsd_start_sector(lp)) * ((off_t) lp->d_secsize);
	if (lseek(fd, offset , 0) != offset)
	{
		wprintw(ewin, "Error: Seek error");
		goto bad_open;
	}

	if (read(fd, buf, BBSIZE) != BBSIZE)
	{
		wprintw(ewin, "Error: read error");
bad_open:
		close(fd);
		wrefresh(ewin);
		return EIO;
	}
	close(fd);

	bh = (void *) buf;
	if (bh->bh_signature != BOOTFLAGS_SIGNATURE)
	{
		wprintw(ewin, "Error: Bad signature");
		wrefresh(ewin);
		return EINVAL;
	}

	while (1)
	{
		WinDisplayBootflags(diskname, lp, &bh->bh_bootflags);

		ans = getresp("what>", 0);
		if (strcmp(ans, "quit") == 0 || strcmp(ans, "exit") == 0 || 
		    *ans == 'q' || *ans == 'x')
			break;

		if (*ans == 0)
		{
			Redraw(diskname, lp);
			continue;
		}

		if (*ans >= '0' && *ans <= '9')
		{
			u_int32_t pos = *ans - '0';
			
			bh->bh_bootflags ^= (1 << pos);
			continue;
		}

		if (strcmp(ans, "edit") == 0)
		{
			WinSwichRaw(diskname, "edit bootflags.conf  ....");
			copy_bootflags(diskname, NULL, 1);
			WinSwichW(diskname, lp);
			continue;
		}

		if (strcmp(ans, "update") == 0)
		{
			install_boot(diskname, lp);
			goto again;
		}

		if (strcmp(ans, "save") == 0)
		{
			ans = getresp("really use kernel settings? (y/n)", 2);
			if (*ans != 'y')
				continue;
			WinSwichRaw(diskname, "save bootflags.conf  ....");
			copy_bootflags(diskname, NULL, 2);
			WinSwichW(diskname, lp);
			continue;
		}

		if (strcmp(ans, "help") == 0 || *ans == '?')
		{
			bootflags_help();
			Redraw(diskname, lp);
			continue;
		}

		if (strcmp(ans, "write") == 0)
		{
			ans = getresp("really write(y/n):", 1);
			if (*ans == 0 || *ans != 'y')
				continue;

			fd = open(devname, O_RDWR, 0);
			if (fd < 0)
			{
				WinDisplayWarn("write open failed");
				continue;
			}

			if (lseek(fd, offset , 0) != offset)
			{
				WinDisplayWarn("lseek failed");
				goto bad_write;
			}

			flag = 1;
			if (ioctl(fd, DIOCWLABEL, &flag) < 0)
			{
				WinDisplayWarn("DIOCWLABEL failed");
				goto bad_write;
			}

			if (write(fd, buf, BBSIZE) != BBSIZE)
			{
				WinDisplayWarn("writing boot failed");
				goto bad_write;
			}

bad_write:
			flag = 0;
			ioctl(fd, DIOCWLABEL, &flag);
			close(fd);
			continue;
		}
	}
	return 0;
}
